//
//  DvString.cc
//
//  DVOS - string class 
//

#include "DvObj.h"
#include "DvString.h"

  int DvString::dblPrecision=16; // static used for all strings when setting from double

  int DvString::getDblPrecision(){
	return dblPrecision;
  }
	
  void DvString::setDblPrecision(int prec){
	dblPrecision = prec;
  }
	
  void DvString::set(double value){
	std::ostringstream buffer;
 // 	buffer.precision(dblPrecision);
	buffer << value;
	
	*this = buffer.str();  	
  }

void DvString::set(double value, int dec){
    
    std::ostringstream buffer;
    if(value == 0.0){
        buffer.setf (std::ios::fixed, std::ios::floatfield);
        buffer.precision(0);
    }
    else if(value < -10000){
        buffer.setf (std::ios::scientific, std::ios::floatfield);
        buffer.precision(dec);
    }
    else if(value < -1){
        buffer.setf (std::ios::fixed, std::ios::floatfield);
        buffer.precision(dec);
    }
    else if(value < 1){
        buffer.setf (std::ios::scientific, std::ios::floatfield);
        buffer.precision(dec);
    }
    else if(value < 10000){
        buffer.setf (std::ios::fixed, std::ios::floatfield);
        buffer.precision(dec);
    }
    else {
        buffer.setf (std::ios::scientific, std::ios::floatfield);
        buffer.precision(dec);
    }
    buffer << value;
    
    *this = buffer.str();
}

DvString DvString::wrapOperator(DvString op){
    DvString first = op.before(' ');
    DvString last = op.after(' ');

    DvString result = first;
    result += *this;
    result += last;
    
    return result;
}


  void DvString::append(double d, const char * spacer){
    *this += spacer;
	ostringstream tmpStr(ostringstream::ate);
	tmpStr.precision(dblPrecision);
	
	tmpStr << d;

  	*this += tmpStr.str();
  }

void DvString::append(double value, int dec, const char * spacer){

    *this += spacer;
    
    std::ostringstream buffer;
    if(value < 1.e-1){
        buffer.setf (std::ios::scientific, std::ios::floatfield);
        buffer.precision(dec);
    }
    else if(value < 10000){
        buffer.setf (std::ios::fixed, std::ios::floatfield);
        buffer.precision(dec);
    }
    else {
        buffer.setf (std::ios::scientific, std::ios::floatfield);
        buffer.precision(dec);
    }
    buffer << value;
    
    *this += buffer.str();
}

 DvString DvString::common(DvString &str){
    
    // copy strings as we'll remove bits
    DvString str1 = *this;
    DvString str2 = str;
    
    // remove mod operators
    str1 = str1.replaceAll("|","");
    str2 = str2.replaceAll("|","");

    // remove non-common blocks in "< >" from both strings
    str1 = str1.cutToken("<", ">", str2);
    str2 = str2.cutToken("<", ">", str1);

    // remove non-common _ to space blocks
    str1 = str1.cutToken("_", " ", str2);
    str2 = str2.cutToken("_", " ", str1);

    // remove non-common space separated blocks
    str1 = str1.cutToken(" ", " ", str2);
    str2 = str2.cutToken(" ", " ", str1);

    // remove non-common space to end blocks
    str1 = str1.cutToken(" ", "", str2);
    str2 = str2.cutToken(" ", "", str1);

    // remove non-common _xx_ name components
    str1 = str1.cutToken("_", "_", str2);
    str2 = str2.cutToken("_", "_", str1);
    
    // step through remaining strings character at a time
    DvString comStr;
    size_t posn = 0;
    size_t posn1 = 0;
    while(posn < str1.size() && posn1 < str2.size()){
        if(str1.at(posn) == str2.at(posn1)) {
            comStr += str1.at(posn);
        }
        posn++;
        posn1++;
    }

    return comStr;
  }

  DvString DvString::cutToken(const char* tok1, const char* tok2, DvString &target){
  
    DvString result = *this;
    DvString sect;
    
    if(tok1[0] == '\0'){
        sect = this->before(tok2);
        sect += tok2;
        if(sect != tok2 && !target.contains(sect)) result = result.replace(sect.c_str(), "");
    }
    else if(tok2[0] == '\0'){
        sect = tok1;
        sect += this->after(tok1);
        if(sect != tok1 && !target.contains(sect)) result = result.replace(sect.c_str(), "");
    }
    else{
        DvString substring = *this;
        sect = this->betweenIncl(tok1, tok2);
        // may be multiple blocks
        while(!sect.empty()){
            if(!target.contains(sect)) result = result.replace(sect.c_str(), " ");// keep space for parsing
            substring = substring.replace(sect.c_str(), " ");
            sect = substring.betweenIncl(tok1, tok2);
        }
    }
    return result;
  }

