

#include "qie_caa.h"

 
int main( int argc, char *argv[] )
{	
	QieData *QieDataCont = new QieData();

	DvString flag;
	int recStart = 0;
	int recEnd = -1;
	int secDecimals = -1;
	bool replace = false;
	bool detached = false;
	bool no_output = false;
	bool compress = false;
 
 	if(argc<2)
	{
		printf("Usage: %s  [-i{tdcpa}] [-o{tdcax}] [-hName] [-nName] [-f] [-v] [-c] [-b] [-sxxx] [-exxx] [-L] [-tn] FileName\n",argv[0]);
		printf("options are...\n");
		printf(" -i{tdcpa} for input type qft, qfd, cdf, pds or cef(CAA) respectively\n");
		printf(" -o{tdcax} to specify output type as  qft, qfd, cdf, cef(CAA), no file respectively\n");
		printf(" -f forces overwrite, otherwise error returned if file exists\n");
		printf(" -v prints version number\n");
		printf(" -V sets verbose feedback\n");
		printf(" -b causes cdf v2.7 files to be created for backward compatibility. Ignored if output not cdf.\n");
		printf(" -hName specifies full Name of a detached header for read, Name.qfh read\n");
		printf(" -nName specifies Name as that of the output file stem\n");
		printf(" -sxxx Records starting with xxx are to be read. First record number is 0. Ignored when reading CDF\n");
		printf(" -exxx Records up to and including xxx are to be read. First record number is 0. Ignored when reading CDF\n");
		printf(" -L turns on writing of end-of-record specifier in header even for default of newline\n");
		printf(" -Pn sets precision in writing double values to n\n");
		printf(" -z turns on compression (gzip) in writing cdf files\n");
		printf(" -tn sets the output accuracy for times to have n places after the decimal point (in seconds)\n");
		printf("     for n > 3 Epoch16 is used in cdf files\n");
		printf("\n");
		printf(" Option flags MUST be space separated,\n");
		printf(" File type option must immediately follow -i and -o flags\n");
 		printf(" Option -h is ignored for cdf type input file\n");
		printf(" If the input file has no extension use -i flag to set input type\n");
		exit(1);
	}

	argc--;
	QieDataCont->inFile.name = argv[argc];

	if(QieDataCont->inFile.name[0] == '-'){
		if( QieDataCont->inFile.name[1] ==  'v') {
			printf("Qtran version %s\n", QIE_VERSION_NUM);
    		exit(0);
  		}
  		else{
    		printf("Option flags must precede input data file\n");
    		exit(1);
  		}
	}

	// remove extension if present and set file type accordingly 
	QieDataCont->inFile.extn = ".";
	QieDataCont->inFile.extn += QieDataCont->inFile.name.afterR('.');
	QieDataCont->inFile.name = QieDataCont->inFile.name.beforeR('.');
	QieDataCont->outFile.name = QieDataCont->inFile.name;
	
	if(QieDataCont->inFile.extn == ".cdf" ) {
		QieDataCont->outFile.extn = ".cef";
	}
	else if(QieDataCont->inFile.extn == ".qft" ) {
		QieDataCont->outFile.extn = ".cdf";
	}
	else if(QieDataCont->inFile.extn == ".qfd" ) {
		QieDataCont->outFile.extn = ".cdf";
	}
	else if(QieDataCont->inFile.extn == ".cef" ) {
		QieDataCont->outFile.extn = ".cdf";
	}
	else if(QieDataCont->inFile.extn == ".tab" ) {
		QieDataCont->outFile.extn = ".cdf";
	}
	else QieDataCont->outFile.extn = ".cef";


	// use arg to override file type 

	while(--argc > 0){
  		flag = argv[argc];
  		if(flag[0] != '-') fprintf(stderr,"unrecognised flag %s\n",flag.c_str());
  		else if (flag[1] == 'o'){
	        if      (flag[2] == 'c') QieDataCont->outFile.extn = ".cdf";
        	else if (flag[2] == 't') QieDataCont->outFile.extn = ".qft";
        	else if (flag[2] == 'd') QieDataCont->outFile.extn = ".qfd";
        	else if (flag[2] == 'a') QieDataCont->outFile.extn = ".cef";
        	else if (flag[2] == 'x') no_output = true;
        }
  		else if (flag[1] == 'i'){
	        if      (flag[2] == 'c') QieDataCont->inFile.extn = ".cdf";
        	else if (flag[2] == 't') QieDataCont->inFile.extn = ".qft";
        	else if (flag[2] == 'd') QieDataCont->inFile.extn = ".qfd";
        	else if (flag[2] == 'a') QieDataCont->inFile.extn = ".cef";
        	else if (flag[2] == 'p') QieDataCont->inFile.extn = ".tab";
        }
  		else if (flag[1] == 'v'){
        	printf("Qtran version %s\n", QIE_VERSION_NUM);
        }
  		else if (flag[1] == 'V'){
        		setDebug(true);
        }
  		else if (flag[1] == 'h'){
	        QieDataCont->inFile.header = &(flag[2]);  
			detached = true;
        }
  		else if (flag[1] == 'n'){
	        QieDataCont->outFile.name = &(flag[2]);
			if(QieDataCont->outFile.name.empty()) QieDataCont->outFile.name = "error";
        }
  		else if (flag[1] == 's'){
	        recStart = atoi(&(flag[2]));  
        }
  		else if (flag[1] == 'e'){
	        recEnd = atoi(&(flag[2]));  
        }
  		else if (flag[1] == 'f'){
	        replace = true;  
        }
  		else if (flag[1] == 'b'){
	          QieDataCont->outFile.CDFbackward = true;
        }
  		else if (flag[1] == 't'){
			secDecimals = atoi(&(flag[2]));
        }
  		else if (flag[1] == 'P'){
			int prec = atoi(&(flag[2]));
			DvString tmp;
			tmp.setDblPrecision(prec); // this sets the global precision in DvString
			QieDataCont->precIsSet = true;
        }
  		else if (flag[1] == 'L'){
			QieDataCont->outFile.force_eol = true;
        }
  		else if (flag[1] == 'z'){
			compress = true;
        }

	}
	  

	///  READ in file  
    //---------------------
	int nRecords = QieDataCont->queryFile(QieDataCont->inFile.name + QieDataCont->inFile.extn);

	printf(" %d Records in file\n", nRecords);
		
	if(recEnd == -1) recEnd = nRecords-1;
	
	if( nRecords > 0) QieDataCont->readData(recStart, recEnd);

	if(no_output) {
		printf(" No output file requested\n");
		exit(0);
	}

	// WRITE new file out, uses same stem name if not explicitly set
	// -------------------------------

	QieDataCont->outFile.attr_delim = ',';
	QieDataCont->outFile.data_delim = ',';
	QieDataCont->outFile.end_of_line = '\n';
	
	// set time resolution in DvTime library
	if(secDecimals >= 0) QieDataCont->setSecDecimals(secDecimals);
   
	if( QieDataCont->outFile.extn == QieDataCont->inFile.extn && QieDataCont->outFile.name == QieDataCont->inFile.name){
		// take care not to overwrite input file 
     
       QieDataCont->outFile.name += "_A";
       printf(" Input and output file names are the same, output modified to %s\n", QieDataCont->outFile.name.c_str());
       
	}
  
	if(QieDataCont->outFile.extn == ".cdf"){
		QieDataCont->writeCDF(replace, compress);   
	}
	else{
		if(QieDataCont->outFile.extn == ".cef") QieDataCont->outFile.end_of_line = '$';
		QieDataCont->writeFlat(replace, detached);
	}

	// reset DvTime library resolution if modified
	if(secDecimals >= 0) QieDataCont->setSecDecimals(QIE_RESET);

	printf( "Done\n" );
	exit(0);
                
}   





