
#include "qie_caa.h"
#include <dirent.h>
#include <iostream>     // std::cout
#include <fstream>      // std::ifstream

using namespace std;

// global 

static const double SecInDay = 86400.; /* we ignore leap seconds */


// ------------------------------------------------------------------------------------
//    QieFT CLASS 
// Free Time format analyser
//    -------------------

QieFT::QieFT()
{	
	
	// free time format elements
	ft_format = "";
	use_ft_offset = false;
	use_DoY = false;
	ft_offset = DvTime();
	ft_all_time = "";
	ft_all_time_fmt = "";
	yy_offset=0;
	mm_offset=0;
	dd_offset=0;
	ss_offset=0.0;
	ft_y=-1; 
	ft_y_len=0;  
	ft_m=-1; 
	ft_m_len=0; 
	ft_m_str=-1; 
	ft_m_str_len=0; 
	ft_d=-1; 
	ft_d_len=0; 
	ft_d_frac=-1; 
	ft_d_frac_len=0; 
	ft_doy=-1; 
	ft_doy_len=0;  
	ft_hr=-1; 
	ft_hr_len=0;  
	ft_hr_frac=-1; 
	ft_hr_frac_len=0;  
	ft_min=-1; 
	ft_min_len=0;  
	ft_min_frac=-1; 
	ft_min_frac_len=0;  
	ft_sec=-1; 
	ft_sec_len=0;  
	ft_sec_frac=-1; 
	ft_sec_frac_len=0;  
	ft_msec=-1; 
	ft_msec_len=0;  
	ft_msec_frac=-1; 
	ft_msec_frac_len=0;  
	ft_musec=-1; 
	ft_musec_len=0;  
	ft_musec_frac=-1; 
	ft_musec_frac_len=0;  
	ft_nsec=-1; 
	ft_nsec_len=0;  
	ft_nsec_frac=-1; 
	ft_nsec_frac_len=0;  
	
	// ok for ft_all_time, resized for actual variable
	year.assign(1,0);
	month.assign(1,0);
	day.assign(1,0);
	dayOfYear.assign(1,0);
	hour.assign(1,0.0);
	minute.assign(1,0.0);
	second.assign(1,0.0);
	
	NdecSec=0;

}


void QieFT::parseFreeTime() {

	if(ft_format.empty()) return;

	NdecSec=0;
	
	/* start looking for key strings in date */
	
	for(size_t ptr=0; ptr<ft_format.size(); ptr++){
		// year
		if(ft_format[ptr] == 'Y'){
			ft_y = ptr;
			if(ft_format[ptr+1] == 'Y'){
				ptr++;
				if(ft_format[ptr+1] == 'Y'){
					ptr++;
					if(ft_format[ptr+1] == 'Y'){
						ptr++;
						ft_y_len = 4;
					}
					else ft_y_len = 3;
				}
				else ft_y_len = 2;
			}
			else ft_y_len = 1;
		}
		else if(ft_format[ptr] == 'M'){
			// month
			if(ft_format[ptr+1] == 'O'){
				if(ft_format[ptr+2] == 'N'){
					ft_m_str = ptr;
					ft_m_str_len = 3;
					ptr+=2;
					if(ft_format[ptr+1] == 'T'){
						ptr++;
						ft_m_str_len++;
						if(ft_format[ptr+1] == 'H'){
							ptr++;
							// Ignore 5th place, 4 characters is enough to get unique month
						}
					}
				}
				else{
					ft_m = ptr;
					ft_m_len = 2;
					ptr++;
				}
			}
			else if(ft_format[ptr+1] == 'M' || ft_format[ptr+1] == 'I'){
				// minute field
				ft_min = ptr;
				ft_min_len = 2;
				ptr++;
				while(ft_format[ptr+1] == 'M' || ft_format[ptr+1] == 'I') {
					ft_min_len++;
					ptr++;
				}
			}
			else if(ft_format[ptr+1] == 'C'){
				// msec field
				ft_msec = ptr;
				ft_msec_len = 2;
				ptr++;
				while(ft_format[ptr+1] == 'C') {
					ft_msec_len++;
					ptr++;
				}
			}
			else {
				// just one M found
				// minute
				ft_min = ptr;
				ft_min_len = 1;
			}

		}
		else if(ft_format[ptr] == 'D'){
			if(ft_format[ptr+1] == 'o' && ft_format[ptr+2] == 'y'){
				// day of year
				ft_doy = ptr;
				ft_doy_len = 3;
				ptr += 2;
				use_DoY = true;
			}
			else if(ft_format[ptr+1] == 'D'){
				// day
				ft_d = ptr;
				ft_d_len = 2;
				ptr++;
				while(ft_format[ptr+1] == 'D') {
					ft_d_len++;
					ptr++;
				}
			}
		}
		else if(ft_format[ptr] == 'H'){
			//hour
			ft_hr = ptr;
			ft_hr_len = 1;
			while(ft_format[ptr+1] == 'H') {
				ft_hr_len++;
				ptr++;
			}
		}
		else if(ft_format[ptr] == 'S'){
			// second
			int ns = -2;
			ft_sec = ptr;
			ft_sec_len = 1;
			while(ft_format[ptr+1] == 'S') {
				ns++;
				ft_sec_len++;
				ptr++;
			}
			if(ns > NdecSec) NdecSec = ns;
		}
		else if(ft_format[ptr] == 'L'){
			// msec
			int ns = 3;
			ft_msec = ptr;
			ft_msec_len = 1;
			while(ft_format[ptr+1] == 'L') {
				ns++;
				ft_msec_len++;
				ptr++;
			}
			if(ns > NdecSec) NdecSec = ns;
		}
		else if(ft_format[ptr] == 'U'){
			// microsec
			int ns = 6;
			ft_musec = ptr;
			ft_musec_len = 1;
			while(ft_format[ptr+1] == 'U') {
				ns++;
				ft_musec_len++;
				ptr++;
			}
			if(ns > NdecSec) NdecSec = ns;
		}
		else if(ft_format[ptr] == 'N'){
			// nanosec
			int ns = 9;
			ft_nsec = ptr;
			ft_nsec_len = 1;
			while(ft_format[ptr+1] == 'N') {
				ns++;
				ft_nsec_len++;
				ptr++;
			}
			if(ns > NdecSec) NdecSec = ns;
		}
		else if(ft_format[ptr] == 'd'){
			// day fractional part
			ft_d_frac = ptr;
			ft_d_frac_len = 1;
			while(ft_format[ptr+1] == 'd') {
				ft_d_frac_len++;
				ptr++;
			}
		}
		else if(ft_format[ptr] == 'h'){
			// hour fractional part
			ft_hr_frac = ptr;
			ft_hr_frac_len = 1;
			while(ft_format[ptr+1] == 'h') {
				ft_hr_frac_len++;
				ptr++;
			}
		}
		else if(ft_format[ptr] == 'm'){
			// minute fractional part
			ft_min_frac = ptr;
			ft_min_frac_len = 1;
			while(ft_format[ptr+1] == 'm') {
				ft_min_frac_len++;
				ptr++;
			}
		}
		else if(ft_format[ptr] == 's'){
			// second, fractional part
			int ns = 1;
			ft_sec_frac = ptr;
			ft_sec_frac_len = 1;
			while(ft_format[ptr+1] == 's') {
				ns++;
				ft_sec_frac_len++;
				ptr++;
			}
			if(ns > NdecSec) NdecSec = ns;
		}
		else if(ft_format[ptr] == 'l' || ft_format[ptr] == 'c'){
			// msec, fractional part
			int ns = 4;
			ft_msec_frac = ptr;
			ft_msec_frac_len = 1;
			while(ft_format[ptr+1] == 'l' || ft_format[ptr] == 'c') {
				ns++;
				ft_msec_frac_len++;
				ptr++;
			}
			if(ns > NdecSec) NdecSec = ns;
		}
		else if(ft_format[ptr] == 'u'){
			int ns = 7;
			// microsec fractional part
			ft_musec_frac = ptr;
			ft_musec_frac_len = 1;
			while(ft_format[ptr+1] == 'u') {
				ns++;
				ft_musec_frac_len++;
				ptr++;
			}
			if(ns > NdecSec) NdecSec = ns;
		}
		else if(ft_format[ptr] == 'n'){
			// nanosec, fractional part
			int ns = 10;
			ft_nsec_frac = ptr;
			ft_nsec_frac_len = 1;
			while(ft_format[ptr+1] == 'n') {
				ns++;
				ft_nsec_frac_len++;
				ptr++;
			}
			if(ns > NdecSec) NdecSec = ns;
		}
		// skip any other character
	
	}

return;

}

void QieFT::assembleTime(size_t n, const char * record){

	if( ft_format == "") return;
	
	// copy bits of record into strings to ensure length used
	
	if(ft_y > -1) {
		string y_str( &(record[ft_y]), ft_y_len );
		year[n] += (int) strtol(y_str.c_str(), NULL, 10);
	}

	if(ft_m > -1) {
		string m_str( &(record[ft_m]), ft_m_len );
		month[n] += (int) strtol(m_str.c_str(), NULL, 10);
	}
	else if(ft_m_str > -1){
		DvString mon_str( &(record[ft_m_str]), ft_m_str_len );
		// DvString does case insensitive comparisons (include french names)
		// 4 digit names may be just first 4 of full name, but are unique
		if(mon_str == "JAN" || mon_str == "JANU" || mon_str == "JANV") month[n] += 1;
		else if(mon_str == "FEB" || mon_str == "FEBR" || mon_str == "FEV" || mon_str == "FEVR") month[n] += 2;
		else if(mon_str == "MAR" || mon_str == "MARC" || mon_str == "MARS") month[n] += 3;
		else if(mon_str == "APR" || mon_str == "APRI" || mon_str == "AVR" || mon_str == "AVRI") month[n] += 4;
		else if(mon_str == "MAY" || mon_str == "MAI") month[n] += 5;
		else if(mon_str == "JUN" || mon_str == "JUNE" || mon_str == "JUIN") month[n] += 6;
		else if(mon_str == "JUL" || mon_str == "JLY"  || mon_str == "JULY" || mon_str == "JUIL") month[n] += 7;
		else if(mon_str == "AUG" || mon_str == "AUGU" || mon_str == "AOU" || mon_str == "AOUT") month[n] += 8;
		else if(mon_str == "SEP" || mon_str == "SEPT") month[n] += 9;
		else if(mon_str == "OCT" || mon_str == "OCTO") month[n] += 10;
		else if(mon_str == "NOV" || mon_str == "NOVE") month[n] += 11;
		else if(mon_str == "DEC" || mon_str == "DECE") month[n] += 12;
		else cout << "Month string not known " << mon_str << endl;
	}
	
	if(ft_d > -1) {
		string d_str( &(record[ft_d]), ft_d_len );
		double d_dbl = strtof(d_str.c_str(), NULL);
		day[n] += (int) d_dbl;
		d_dbl -= (int) d_dbl;
		second[n] += d_dbl * SecInDay;

	}
	if(ft_d_frac > -1) {
		string df_str = string(".") + string(&(record[ft_d_frac]), ft_d_frac_len);
		double df = strtof(df_str.c_str(), NULL) * SecInDay;
		second[n] += df;
	}
	
	if(ft_doy > -1) {
		string doy_str( &(record[ft_doy]), ft_doy_len );
		dayOfYear[n] += (int) strtol(doy_str.c_str(), NULL, 10);
	}
	
	if(ft_hr > -1) {
		string hr_str( &(record[ft_hr]), ft_hr_len );
		hour[n] += strtof(hr_str.c_str(), NULL);
	}
	if(ft_hr_frac > -1) {
		string hf_str = string(".") + string(&(record[ft_hr_frac]), ft_hr_frac_len);
		double hf = strtof(hf_str.c_str(), NULL) * 3600.0;
		second[n] += hf;
	}
	
	if(ft_min > -1) {
		string min_str( &(record[ft_min]), ft_min_len );
		minute[n] += strtof(min_str.c_str(), NULL);
	}
	if(ft_min_frac > -1) {
		string mf_str = string(".") + string(&(record[ft_min_frac]), ft_min_frac_len);
		double mf = strtof(mf_str.c_str(), NULL) * 60.0;
		second[n] += mf;
	}
	
	if(ft_sec > -1) {
		string sec_str( &(record[ft_sec]), ft_sec_len );
		second[n] += strtof(sec_str.c_str(), NULL);
	}
	if(ft_sec_frac > -1) {
		string sf_str = string(".") + string(&(record[ft_sec_frac]), ft_sec_frac_len);
		second[n] += strtof(sf_str.c_str(), NULL);
	}
	
	if(ft_msec > -1) {
		string msec_str( &(record[ft_msec]), ft_msec_len );
		second[n] += strtof(msec_str.c_str(), NULL) * 0.001;
	}
	if(ft_msec_frac > -1) {
		string msf_str = string(".") + string(&(record[ft_msec_frac]), ft_msec_frac_len);
		second[n] += strtof(msf_str.c_str(), NULL) * 0.001;
	}
	
	if(ft_musec > -1) {
		string musec_str( &(record[ft_musec]), ft_musec_len );
		second[n] += strtof(musec_str.c_str(), NULL) * 0.000001;
	}
	if(ft_musec_frac > -1) {
		string musf_str = string(".") + string(&(record[ft_musec_frac]), ft_musec_frac_len);
		second[n] += strtof(musf_str.c_str(), NULL) * 0.000001;
	}
	
	if(ft_nsec > -1) {
		string nsec_str( &(record[ft_nsec]), ft_nsec_len );
		second[n] += strtof(nsec_str.c_str(), NULL) * 0.000000001;
	}
	if(ft_nsec_frac > -1) {
		string nsf_str = string(".") + string(&(record[ft_nsec_frac]), ft_nsec_frac_len);
		second[n] += strtof(nsf_str.c_str(), NULL) * 0.000000001;
	}

	return;
	
}

// ------------------------------------------------------------------------------------
//    QieVariable CLASS 
//    -------------------

QieVariable::QieVariable(const char *variableName)
{	
	name = variableName;
	col_start = -1; // unused for cdf
	col_width = -1; // unused for cdf
	recVary = true;
	datatype = DATA_STR;
	n_posn = -1; // unused for cdf
	objDim.clear();
	nRecs = -1;
	
	// free time format elements
	useFreeFormat = false;
	skip = false;	
    
    // create dummy variable to be replaced (holds xrefs until var ready)
    Vobj = new DvObject();
}

QieVariable::~QieVariable()
{
}



//    -------------------

void QieVariable::processTime(){
	// convert free time
	int year=0;
	int month=0;
	int day=0;
	int dayOfYear=0;
	int hour=0;
	int minute=0;
    int sec=0;
    int nano=0;
	double second=0.0;

	// start with offset
	if(ft.use_ft_offset) ft.ft_offset.breakDownTAI(&year, &month, &day, &hour, &minute, &sec, &nano );
	second = (double)sec;
    second += (double)nano * 1.e-9;
    
	year += ft.yy_offset;
	month += ft.mm_offset;
	if(ft.use_DoY) dayOfYear += ft.dd_offset;
	else day += ft.dd_offset;
	second += ft.ss_offset;

	if(!ft.ft_all_time_fmt.empty() && !ft.ft_all_time.empty()){
		QieFT all_time;
		all_time.ft_format = ft.ft_all_time_fmt;
		// turn format into ptr offsets and length
		all_time.parseFreeTime();
		
		// convert time string to elements using ptrs and lengths
		// single value fits into element zero of default vectors of 1 element
		all_time.assembleTime(0, ft.ft_all_time.c_str());
		
		year += all_time.year[0];
		month += all_time.month[0];
		day += all_time.day[0];
		second += all_time.hour[0]*3600.;
		second += all_time.minute[0]*60.;
		second += all_time.second[0];
	}

	if(ft.use_DoY) {
		for(size_t i=0; i<nEntries*nRecs; i++){
			double sec_i = ft.second[i]+second;
			int hr_i = (int) ft.hour[i];
			int min_i = (int) ft.minute[i];
			sec_i += 3600. * (ft.hour[i] - (double) hr_i); // add fractional part
			sec_i += 60. * (ft.minute[i] - (double) min_i); // add fractional part
			
			Vobj->time(i).setFromDOY(ft.year[i]+year, ft.dayOfYear[i]+dayOfYear, hr_i+hour, min_i+minute, sec_i );
		}
	}
	else{
		for(size_t i=0; i<nEntries*nRecs; i++){
			double sec_i = ft.second[i]+second;
			int hr_i = (int) ft.hour[i];
			int min_i = (int) ft.minute[i];
			sec_i += 3600. * (ft.hour[i] - (double) hr_i); // add fractional part
			sec_i += 60. * (ft.minute[i] - (double) min_i); // add fractional part

			Vobj->time(i).setFromUT(ft.year[i]+year, ft.month[i]+month, ft.day[i]+day, hr_i+hour, min_i+minute, sec_i);
		}
	}

	Vobj->secRes = ft.NdecSec;
}

// -------------------------------------------


void QieVariable::ensureDeltas(){

    if(Vobj.is_nil()) return;

    DvNode *ptrDPnode  = Vobj->find_xref(DELTA_PLUS_VAR);
    DvNode *ptrDMnode  = Vobj->find_xref(DELTA_MINUS_VAR);

    DvNode *DPnode  = Vobj->find_xref(DELTA_PLUS);
    DvNode *DMnode  = Vobj->find_xref(DELTA_MINUS);

    // rename DELTA_PLUS_VAR to DELTA_PLUS etc
    if(!DPnode && ptrDPnode) {
        ptrDPnode->setName(DELTA_PLUS);
        DPnode = ptrDPnode;
    }
    if(!DMnode && ptrDMnode) {
        ptrDMnode->setName(DELTA_MINUS);
        DMnode = ptrDMnode;
    }

    if(Vobj->is_time()){
        // Ensure we have Delta_plus and Delta_minus for timelines and clean up
	
        DvNode *SICnode = Vobj->find_xref(SI_CONVERSION);
        DvNode *UNITSnode = Vobj->find_xref(UNITS);

        // DvTime is in seconds (conversion already done on read)
        if(SICnode) {
            DvObject_var SIc = new DvObject( "1>s" );
            SICnode->setObj(SIc); 
        }
        if(UNITSnode){ 
            DvObject_var unit = new DvObject( "s" );
            UNITSnode->setObj(unit);
        }
        
        if(DPnode && DMnode) return;
	
	
        double spacing = Vobj->get_spacing();
	
        if(DMnode && DMnode->obj().is_ok() ){
            spacing -= DMnode->obj()->asDouble();
            DvObject_var delta = new DvObject(spacing);
            if(Vobj->is_time()) delta->change_xref(SI_CONVERSION, "1>s");
            else {
                if(SICnode) delta->change_xref(SI_CONVERSION, SICnode->obj() );
            }
            Vobj->change_xref(DELTA_PLUS, delta);
        }
        else if(DPnode && DPnode->obj().is_ok()){
            spacing -= DPnode->obj()->asDouble();
		
            DvObject_var delta = new DvObject(spacing);
            if(Vobj->is_time()) delta->change_xref(SI_CONVERSION, "1>s");
            else {
                if(SICnode) delta->change_xref(SI_CONVERSION, SICnode->obj() );
            }
            Vobj->change_xref(DELTA_MINUS, delta);
        }
        else{
            DvObject_var delta = new DvObject(spacing*0.5);
            if(Vobj->is_time()) delta->change_xref(SI_CONVERSION, "1>s");
            else {
                if(SICnode) delta->change_xref(SI_CONVERSION, SICnode->obj() );
            }
            Vobj->change_xref(DELTA_PLUS, delta);
            Vobj->change_xref(DELTA_MINUS, delta);
        }
    }
	return;
}




// -----------------------------------------------------------------------------------------
//    QieFile CLASS  
//    -------------------


/// \fn QieFile::QieFile()
/// \brief Constructor of the QieFile class
/// \param     none
/// \exception None
/// \return    None
///  

QieFile::QieFile()

{
  start_after = std::string();
  start_with = std::string();
  data_until = std::string();
  name = "";
  extn = "";
  
   // set defaults
  end_of_line = '\n';
  commentMark = '!';
  attr_delim = ',';
  data_delim = ',';
  content = NULL;
  hasTitle=false;
  CDFbackward=false;
  force_eol=false;
  nToSkip=0;

}

// ------------------------

QieFile::~QieFile()

{
	this->closeFile();

}


// -------------------------------------------

DvString QieFile::safePath(const char * path)
{
	DvString safepath;
	DvString pathcopy;
	char *home = NULL;
	char pathDelim = '/';
	DvString desktop("~/Desktop/");

    pathcopy = path;
	if(path[0] == pathDelim){
		// ensure this is not root
		if( !pathcopy.after(pathDelim).contains(pathDelim) ) {
			// in root dir, move it
			pathcopy.replace(DvString(pathDelim).c_str(), desktop.c_str()); // move to Home Desktop
		}
	}
	
	if(pathcopy[0] == '~'){
		if(pathcopy[1] == pathDelim ){
       		if ( (home = getenv("HOME")) != NULL ){
         		safepath = home;
       		}
       		else{
				safepath = ".";
       		}
    	}
		else safepath = "."; // else place wherever QSAS/Qtran was launched from

		
    	safepath += pathDelim;
		DvString tail = pathcopy.after(pathDelim);
    	if(tail.empty()) safepath += "export";
    	else safepath += tail;

  	}
	else{   /* nothing to expand */
    	safepath = pathcopy;
	}

	return safepath;

} 


 
// --------------------------------

void QieFile::openFile(DvString fullName){
	
	if(fullName.empty()) return;

	this->name = fullName.c_str();
	
	// clear eof flag or any errors
	ifp.clear();
	
	// try reading to see if file of that name exists 
	// use binary to stop Windows msys trying to remove \r and messing up
	ifp.open(this->name.c_str(), std::fstream::in | std::fstream::binary);
	
    if(ifp.is_open()){

        // get length of file:
        ifp.seekg (0, ifp.end);
        int length = ifp.tellg();
        ifp.seekg (0, ifp.beg);

        // allocate memory:
	    if(content) delete[] content;

        content = new char[length+2];
	
        // read data as a block:
        ifp.read(content, length);

	    posn = 0;

	    content[length] = '\0';
	    content[length+1] = '\0'; // needed if last line in file is comment
    }
}

// -----------------------------------------------------

size_t QieFile::countRecords(){
		
	if(content == NULL || strlen(content) <= posn) return 0;
	
    char c='1'; // can be anything here
    
     while(nToSkip>0 && c != '\0'){
        c = skipRestOfLine();
        nToSkip--;
    }

	recPosn.clear();
	char * next = &(content[posn]);
	// skip n lines for '&n' flag
	if(*next == '&'){
		next++;
		int nToSkip = (int) strtol(next, NULL, 10);
		while(nToSkip>=0){ // skip n lines plus this one
			next++;
			while(*next != '\n' && *next != '\0') next++;
			if(*next == '\n') next++;
			nToSkip--;
		}
	}
    // skip to after start_after
    if( !start_after.empty()) {
        char *start = strstr(content, start_after.c_str());
        if (start != NULL) next = start;
        while(*next != '\n' && *next != '\0') next++; // get rid of this line
        if(*next == '\n') next++;
    }
    
    // skip to after start_with
    if( !start_with.empty()) {
        char *start = strstr(content, start_with.c_str());
        if (start != NULL) next = start;
    }
    
	// skip commented lines at start
	while(*next == commentMark) {
		next++;
		while(*next != '\n' && *next != '\0') next++;
		if(*next == '\n') next++;
	}

	if(hasTitle){
		// skip first line
		next++;
		while(*next != '\n' && *next != '\0') next++;
		if(*next == '\n') next++;
	}
	
	char * recStart = next;
	while(*next != '\0') {
		// skip commented lines
		while(*next == commentMark && *next == *recStart) {
			// comment line at start (don't care about comments tagged on end)
			next++;
			while(*next != '\n' && *next != '\0') next++; // find end of line
			while(*next == '\n' || *next == '\r' ) next++; // step over it
			recStart = next;
		}
		
		if(*next == end_of_line) {
			recPosn.push_back(recStart);
			recStart = next+1;
			if(end_of_line != '\n'){
				// remove trailing space
				while (*recStart == ' ' ) recStart++;
			}
			while (*recStart == '\n' || *recStart == '\r' ) recStart++;
		}
		next++;		
	}
	recPosn.push_back(recStart);

	if(recPosn.size() == 0) return 0;
	
	while( *(recPosn[recPosn.size()-1]) == '\0' ) recPosn.pop_back();

	if( !data_until.empty() && strncmp(recPosn[recPosn.size()-1], data_until.c_str(), data_until.size()) == 0 ) recPosn.pop_back();
	
	return recPosn.size();
}


// -----------------------------------------------------
// 
// returns true if something was read OK otherwise false
// Leaves ifp ready to read next record
 
bool QieFile::skipNextRecord(){
	
	char c;
	
	// read from this File ptr
	if(!ifp.is_open()) return false;
	c = get();
	
	// ignore commented line
	while(c == '!') {
		c=skipRestOfLine();	
		return skipNextRecord();
	}
	
    // skip empty lines
	while (c == ' '  || c == '\t' || c == '\n' || c == '\r') {
		if(c == '\0') return false;  // ran out of data without finding end of line
		c = get();
	}
	
    // start of non-commented, non-empty line
	while (c != end_of_line ){
		if(c == '\0') return false;  // ran out of data without finding end of line
		c = get();	
    }
	return true;
}



// ------------------------------------------------------

char QieFile::readLine(size_t &posStart, size_t &posEnd){
	
	// locates input lines up to '\n' and concatenates on continuation marker
    
	// Adds one space at the end to make concatenation safe.

	// endMark may be '\n' or end of record

	char c = get();
	
	if(c == '&'){
		nToSkip = (int) strtol(&(content[posn]), NULL, 10) + 1; // + to skip this line
	}
    
    while(nToSkip>0 && c != '\0'){
        c = skipRestOfLine();
        nToSkip--;
    }

	// remove leading white space and empty lines
	while(c != '\0')
	{
		if(c == ' ' || c == '\n'|| c == '\r' || c == '\t') c = get();
		else if(  c == commentMark) {
			// skip comment/blank lines
			c = skipRestOfLine();
		}
		else break;
	}

	// there is now something in the current line
	posStart = posn-1;
	
	while(c != '\n' &&  c != '\0'){
	
		if( c == '\r') {
			// skip new line 
			content[posn-1] = ' ';
			c = get();
		}
		else if( c == '\\') {
            c = get();
            if( c == '\\') {
			    content[posn-1] = ' ';// remove "\"
			    // skip comment or continuation mark to end of line ("\")
			    c = skipRestOfLine();
			    if(c == '\n'){
				    // blank out line end
				    content[posn-1] = ' ';
				    c = get();
			    }
            }
		}
		else if( c == commentMark) {
			// skip comment after valid line 
			posEnd = posn-1;
			c = skipRestOfLine();
			return c; // there is something in this line
		}
		else if(c == '\"') {
			// take quoted text verbatim
			c = get();
			while( c != '\"'){
				c = get();

//				if( c == '\0' || c == end_of_line) {
				if( c == '\0' ) {
					// unclosed quote
					// string may be rest of file, so truncate
					char cutText[21];
					strncpy(&(cutText[0]), &(content[posStart]), 20);
					QieMsg("Error",  "Unclosed quotes after line starting... \n", &(cutText[0]), "\n");
					posEnd = posn-1;
                    
					return c; 
				}
			}
			c = get();
		}	
		else if(c == end_of_line) {
			posEnd = posn-1;
			return c;
		}
		else{
			// add this character and test next
			c = get();
		}					
	}
	posEnd = posn-1;
	if(posEnd > posStart) c = '\n'; // something found
	
	return c; 
 }
 
// ------------------------------------------------------

char QieFile::get(){
	char c = content[posn];
	posn++;
	return c;
}


// -------------------------------------------------------

int QieFile::skipRestOfLine(){
	// ignore input up to the end of line (or file if sooner)
	
	int c = get();
	while(c != '\n' &&  c != '\0') {
		content[posn-1] = ' '; // blank out 
		c = get();
	}
	return c;
}

// -------------------------------------------------------

void QieFile::closeFile(){

	if(extn == ".CDF") 
	{
		CDFlib(CLOSE_, CDF_, NULL_);  // safe against multiple closes
		return;
	}
	
	// trap multiple close on flat files

	if (ifp.is_open()) ifp.close();
	if (ofp.is_open()) ofp.close();

	if (content) delete[] content;	
    content = NULL;
	
}


// -------------------------------------------------------

void QieFile::setEOL(char eol){
	end_of_line = eol;
}

// -------------------------------------------------------

void QieFile::setCMark(char cmark){
	commentMark = cmark;
}

// -------------------------------------------------------

void QieFile::setStartAfter(DvString &txt){
    start_after = txt.c_str();
}

// -------------------------------------------------------

void QieFile::setStartWith(DvString &txt){
    
    start_with = txt.between('\"', '\"').c_str();
}

// -------------------------------------------------------

void QieFile::setDataUntil(DvString &txt){
	data_until = txt.c_str();
}

// -------------------------------------------------------

void QieFile::setAttrDelim(char c){
	attr_delim = c;
}

// -------------------------------------------------------

void QieFile::setDataDelim(char c){
	data_delim = c;
}

// ----------------------------------------------------------
bool QieFile::dataStartIn(size_t posStart, size_t posEnd){
 
	static char c = start_after[0];
	for(size_t i=posStart; i<=posEnd; i++){
		if(content[i] == c){
			bool found = true;
			for(size_t j=1; j<start_after.size(); j++) {
				i++;
				if(content[i] != start_after[j]) {
					found = false;
					break;
				}
			}
			if(found) return true;
		}
	}
	return false;
}


// ------------------------------------------------------------------------------------
//    QieData CLASS 
//    -------------------


/// \fn QieData::QieData()
/// \brief Constructor of the QieData class
/// \param     none
/// \exception None
/// \return    None
///  

QieData::QieData()
{
	// start with empty variable and global lists
	QieVars.clear();
	recordNumEntries = 0;
	d0_var_num = -1;
    d0_var_len = -1;
	nRecords = 0;
	dataSecRes = -1;
	precIsSet = false;
}

QieData::~QieData()
{
	for(size_t i=0; i<QieVars.size(); i++) {
		delete QieVars[i];
	}

	inFile.closeFile(); 
	outFile.closeFile(); 
	QieVars.clear();

}



// ----------------------------------------------------------

int QieData::queryFile(DvString fullName)
{
	bool usingHeader=false;

	if(fullName.empty()) return 0;
	
	if(fullName.size() < 5) return 0;

	// initialise guess of file type if not set by calling routine
	if(inFile.extn.empty()){
		inFile.extn = fullName.substr(fullName.size()-4, 4);
	}
	
	bool header_OK=false;

	// branch to cdf reading options

	if(inFile.extn == ".CDF") {
		nRecords = queryCDFfile(fullName);
		if(nRecords == 0) QieMsg ("WARNING", "No data in file or unable to read cdf file\n");
		return nRecords;
	}

	if(inFile.extn == ".TAB") {
		nRecords = queryTABfile(fullName);
		if(nRecords == 0) {
            QieMsg ("WARNING", "No data in file or .LBL file missing, trying tabular\n");
            inFile.extn = ".qft";
        }
		else return nRecords;
	}
	
	// default is CAA syntax cef file
	// first read detached header if provided
	
	if( !inFile.header.empty()){
		inFile.openFile(inFile.header.c_str());
		usingHeader = true;
	}
	else inFile.openFile(fullName);
	
	if(inFile.ifp.is_open()){
		header_OK = readHeader();
		if( !header_OK ) QieMsg ("WARNING", "No header in file, looking for detached header\n");
	}
	
	if( !header_OK ) {

		// search for detached header
		vector <DvString> hdrList = QieFindHelperFiles(fullName.c_str(), ".qfh");
		DvString hdrName;
		
		// cannot tell which header is applicable, so only use first found
		if(hdrList.size() > 0) hdrName = hdrList[0]; 
		else hdrName = "";

		if ( !hdrName.empty() ){
			usingHeader = true;
			inFile.closeFile(); // close data file and open detached header
			inFile.openFile(hdrName);
			if (inFile.ifp.is_open()) {
				header_OK = readHeader();
				if( !header_OK ){
					QieMsg ("ERROR", "Cannot read header\n", hdrName.c_str());
					return 0;
				}
			}
		}
		else{
			QieMsg ("ERROR", "Cannot locate a valid header\n");
			inFile.closeFile(); // close data file
			return 0;
		}	
	}
	
	if( usingHeader) {
		inFile.closeFile();	
	}
	
	// Add attributes from helper files
	getAttributeHelper(fullName.c_str());
		
	if( usingHeader) {
		inFile.openFile(fullName); // 
	}

	nRecords = inFile.countRecords();

    return nRecords;

}


// ----------------------------------------------------------

void QieData::readData(size_t recStart, size_t recEnd){
	// reads all variables (for Qtran)

	// create list of vars to fetch and initialise
	vector <size_t> getList;
	for(size_t i=0; i<QieVars.size(); i++){
				
		if(inFile.extn == ".CDF" ){
			// get all vars (NRV not done in queryCDFfile())
			getList.push_back(i);
		}
		else if( inFile.extn == ".TAB"  && QieVars[i]->recVary){
			// process all vars (already read in in qeuryTABfile())
			getList.push_back(i);
		}
		else if( QieVars[i]->recVary){
			// excludes NRV vars for flat files as read in header
			initSequence(QieVars[i], recEnd - recStart + 1);
			getList.push_back(i);
		}
	}
	
	if(getList.size() == 0) return;
	
	if(inFile.extn == ".CDF") {
	   // fetch data for variables from CDF file
	   getCDFvarDataList(getList, recStart, recEnd);
	   nRecords = recEnd - recStart + 1;
	   return;
	}
	else if (inFile.extn == ".QFT" ){

		// get records requested for tabular data
		int recNum=0;
		
		for( size_t recNow=recStart; recNow<=recEnd; recNow++ ){
			// locate start and length of each token in record
		
			nEntry = -1; // number of entry in record
			recVarEntry = inFile.recPosn[recNow]; // location of start of record
			for(size_t nv=0; nv<getList.size(); nv++){
				getVarRecordTableEntry(getList[nv], recNum);
			}

			recNum++;
		
		} // end while over records

	    // The variables needed have now been read into their Vobj pointers for the records requested
	}
	else if (inFile.extn == ".TAB"){

		// get records requested for Rosetta data
		// data was already read in on queryTABfile()
		if(recStart != 0 || recEnd != nRecords-1) {
			// subset
			for( size_t i=0; i<getList.size(); i++ ){
				QieVariable *V = QieVars[getList[i]];
				if(V->Vobj->seqSize() == nRecords) V->Vobj = V->Vobj->subSequence(recStart, recEnd);
			} 
	    }
		
	}
	else if(inFile.data_delim == ' '){
		// get records requested for appropriate data type
		int recNum=0;
		
		for( size_t recNow=recStart; recNow<=recEnd; recNow++ ){
			// locate start and length of each token in record
		
			nEntry = -1; // number of entry in record
			recVarEntry = inFile.recPosn[recNow]; // location of start of record
			for(size_t nv=0; nv<getList.size(); nv++){
				getVarRecordEntrySpSep(getList[nv], recNum);
			}

			recNum++;
		
		} // end while over records

	    // The variables needed have now been read into their Vobj pointers for the records requested
	}
	else{
		// get records requested for appropriate data type
		size_t recNum=0;
		for( size_t recNow=recStart; recNow<=recEnd; recNow++ ){
			// locate start and length of each token in record
		
			nEntry = -1; // number of entry in record
			recVarEntry = inFile.recPosn[recNow]; // location of start of record
			for(size_t nv=0; nv<getList.size(); nv++){
				getVarRecordEntry(getList[nv], recNum);
			}

			recNum++;
		
		} // end while over records

	    // The variables needed have now been read into their Vobj pointers for the records requested
	}

	// post process Free Time variables to assemble complete time
	// Also ensure it has Deltas
	for( size_t i=0; i<getList.size(); i++ ){
		QieVariable *V = QieVars[getList[i]];
		if(V->datatype == DATA_TIME){
			if( V->useFreeFormat){
             
                V->processTime();
                
				// first partial time var is the one we attach
				if(mergeTimeVar(i)) break; // only one broken up time possible
				
			}
			else if(V->Vobj->totalSize() > 0){
				V->Vobj->secRes = V->Vobj->getSecResolution();
			}
			V->ensureDeltas();
		}
	}

	nRecords = recEnd - recStart + 1;

	return;

}

// ----------------------------------------------------------

void QieData::getAttributeHelper(const char * fullName)
{
	vector <DvString> helperList = QieFindHelperFiles(fullName, ".qat");

	for (size_t i=0; i<helperList.size(); i++){
		const char *helperName = (helperList[i]).c_str();

		if ( !QieStrIsNULL(helperName) ){
			helperFile.openFile(helperName);
			if (helperFile.ifp.is_open()) {
				QieMsg ("INFO:", helperName, "being read for additional attribute information\n");
				readAttrHelper();
			}
		}
	}

	return;
} 

// ----------------------------------------------------------

bool QieData::notHomogeneous(){
    
    bool notHomog = false;
    
    for(size_t i=0; i<QieVars.size(); i++){
        QieVariable *V = QieVars[i];
        if(V->recVary && V->Vobj->seqSize() != nRecords) notHomog = true;
    }
    
    return notHomog;
    
}

// ----------------------------------------------------------

void QieData::readAttrHelper()
{
 	// CAA defaults may not be written in the file, so set in advance
	helperFile.attr_delim = ',';
	helperFile.data_delim = ',';
	vector <QieVariable*>  unused; 
	
	// note: readLine skips comment lines 
	size_t posStart;
	size_t posEnd;

	while(helperFile.readLine(posStart, posEnd) != '\0'){
		
		if( !parseLine(helperFile.content, posStart, posEnd) ){
			// detect (usually) data lines with no header or Start_data
			if( isdigit(parameter[0]) ) break; // non-fatal error
			else QieMsg( "WARNING", "unidentified input line ", string(&(helperFile.content[posStart]), posEnd - posStart + 1).c_str());
			continue;
		}
		
		if(parameter == "START_VARIABLE") {
			DvString name = value.clean();
			int vnum = findVarNumByName(name.c_str());
			
			QieVariable *V;
			if(vnum < 0){
				// New variable not in real header 
				V = this->initVariable(name.c_str(), &helperFile);
				if(V) unused.push_back(V);
			}
			else{
			
				// read attribute info for this variable
				while(helperFile.readLine(posStart, posEnd) != '\0'){
					if( !parseLine(helperFile.content, posStart, posEnd) ){
						QieMsg( "WARNING", "unidentified input line ", string(&(helperFile.content[posStart]), posEnd - posStart + 1).c_str());
						continue;
					}

					if(parameter == "END_VARIABLE") {
						if(name != value.clean()) {
							cout << "Variable Block in helper file, start and end names differ";
						}
						break;
					}
				
				
					V = QieVars[vnum];		
				
					// read attribute 
					
					if(parameter == "TIME_FORMAT") {
						if(value.clean() == "FREE_TIME_FORMAT") {
   		       				V->useFreeFormat = true;
						}
						else if(value.clean() == "ISO"){
     	       				V->useFreeFormat = false;
						}         
						else{
     	       				V->useFreeFormat = false;
							QieMsg( "WARNING", "Unknown time format designation '", value.cstr(), "' being ignored");
						}
						continue;
					}

					if(parameter == "TIME_FORMAT_STRING") {
						V->ft.ft_format = value.cstr();
						V->ft.parseFreeTime();
          				continue;
					}

					
					if(parameter == "SIZES") {
						
						int n_elems = countEntryElements(value, helperFile.attr_delim);
						// set up Dim object 
			
						if(n_elems == 0)  continue;  
		
						V->objDim.clear();
						V->objDim.assign(n_elems, 0);	
						int startAt = 0;
						string token;
						for(int i=0; i<n_elems; i++){
					
							token = getNextEntry(value, startAt, helperFile.attr_delim);
							int Ndim = (int) strtol(token.c_str() , NULL, 10);
							if(Ndim == 0) {
								QieMsg( "ERROR", "Variable dimension size zero from '", value.c_str());
								return;
							}
							V->objDim[i] = Ndim;
						}
						continue;
					}

					// else read attribute
					DvObject_var nextAttr = this->QieText_to_Attr(value, &helperFile); 
					
					if( nextAttr.is_ok() ){
						// REPLACE that in the file if present
						V->Vobj->change_xref(parameter.c_str(), nextAttr);
					}
					
				}
			}
		}
		
		if(parameter == "COMMENT_MARKER") {
            helperFile.setCMark(QieGetDelim(value));
            inFile.setCMark(QieGetDelim(value));
			continue;
		}
		if(parameter == "ATTRIBUTE_DELIMITER") {
            helperFile.setAttrDelim(QieGetDelim(value));
            inFile.setAttrDelim(QieGetDelim(value));
			continue;
		}
		if(parameter == "DATA_DELIMITER") {
            helperFile.setDataDelim(QieGetDelim(value));
            inFile.setDataDelim(QieGetDelim(value));
			continue;
		}	

		
	} // end while


 	// add variables matching this file
	for(int i=unused.size()-1; i>=0; i--){
		bool keep = false;
		DvNode *XnodeUseWith = unused[i]->Vobj->find_xref("Use_With");
		if(XnodeUseWith) {
		
			DvString useWithName = XnodeUseWith->obj()->asStr();
			if(useWithName == "Always" || findVarNumByName(useWithName.c_str()) >= 0){
				QieVars.push_back(unused[i]);
				keep = true;
			}
			unused[i]->Vobj->delete_xref("Use_With");

		}
		if(!keep) delete unused[i];
	}
	helperFile.closeFile();
}

// ----------------------------------------------------------

bool QieData::readHeader()
{
	bool varFound=false;
	char endOfRecord = '\n';

 	// CAA defaults may not be written in the file, so set in advance
	inFile.attr_delim = ',';
	inFile.data_delim = ',';

	// note: readLine skips comment lines 
	size_t posStart;
	size_t posEnd;
	
	while(inFile.readLine(posStart, posEnd) != '\0'){
		if( !inFile.start_after.empty() ){
			// skip until start after line found
			if(inFile.dataStartIn(posStart, posEnd)) {
				QieMsg( "INFO", "Start reading data after line =", inFile.start_after.c_str());
				inFile.start_after = "";
			}
			continue; // keep looking for start_after line
		}
		
		if( !parseLine(inFile.content, posStart, posEnd) ){
			// detect (usually) data lines with no header or Start_data
			//if( isdigit(parameter[0]) ) break; // non-fatal error
		   // else QieMsg( "WARNING", "unidentified input line ", nextLine.c_str());
			break;
		}
		
		if(parameter == "FILE_FORMAT_VERSION" ) {
			if( value.clean() == "CEF-2" && inFile.extn.empty()) inFile.extn = ".cef";
			continue;
		}
	  
        if(parameter == "START_DATA") {
            QieMsg("INFO", value.c_str(), "records expected");
            break;
        }
        
		if(parameter == "DATA_UNTIL") {
			if(inFile.extn.empty()) inFile.extn = ".cef";
			inFile.setDataUntil(value.clean());			
			break;
		}
		if(parameter == "START_VARIABLE") {
			QieVariable *V = this->initVariable(value.cstr());
			if(V){
				QieVars.push_back(V);

				varFound = true;
				continue;
			}
			else return false;
		}
		if(parameter == "START_META") {
			this->readGlobal(value.cstr());
			continue;
		}
		if(parameter == "FILE_TYPE") {
			if(value[0] == 't' || value[0] == 'T' ){
				if(inFile.extn != ".CAA") inFile.extn = ".qft";
			}
			else if(value[0] == 'd' || value[0] == 'D' ){
				if(inFile.extn != ".CAA") inFile.extn = ".qfd";
			}
			continue;
		}
		if(parameter == "COMMENT_MARKER") {
			inFile.setCMark(QieGetDelim(value));
			continue;
		}
        if(parameter == "START_AFTER") {
            inFile.setStartAfter(value.clean());
            continue;
        }
        if(parameter == "START_WITH") {
            inFile.setStartWith(value);
            continue;
        }
		if(parameter == "ATTRIBUTE_DELIMITER") {
			inFile.setAttrDelim(QieGetDelim(value));
			continue;
		}
		if(parameter == "DATA_DELIMITER") {
            char c = QieGetDelim(value);
			inFile.setDataDelim(c);
			continue;
		}	
		if(parameter == "END_OF_RECORD_MARKER") {
			if (value[0] != ' ') {
				value.clean();
				if( value[0] == '\n' || value[0] == '\0') 
					endOfRecord = '\n';
				else 
					endOfRecord = QieGetDelim(value);
			}
			continue;

		}
		if(parameter == "FILE_NAME") {
			QieMsg("INFO", "file: ", value.c_str());
			continue;
		}

	} // end while

	// validate cef 2 inputs found 
	if(inFile.extn == ".cef"){
		if( inFile.start_after != std::string("") )
			QieMsg( "WARNING", "START_AFTER parameter should not be used in CAA file. ", " Continuing to read");
		if(inFile.commentMark != '!')
			QieMsg( "WARNING", "CEF 2 format: Comment marker not exclamation mark '!'. ", " Continuing to read" );
		if(inFile.attr_delim != ',')
			QieMsg( "WARNING", "CEF 2 format: Attribute delimiter must be a comma. ", " Continuing to read" );
		if( inFile.data_delim != ',')
			QieMsg( "WARNING", "CEF 2 format: Data delimiter must be a comma. ", " Continuing to read" );
	}

	inFile.setEOL( endOfRecord ); // only set this AFTER the header has been read as only records use end_of_line
	//   QieFTinit_parsers();  // initalise parsers for SJS free time routines 

	// ensure there is a DEPEND_0 set if time line exists
	ensureDependZero();

	return varFound;
} 

// ------------------------------------------------

int QieData::getDataType(DvString dataType)
{
	if(dataType == "DOUBLE" ) return DATA_DBL;
	else if(dataType == "FLOAT") return DATA_FLT;
	else if(dataType == "INT" || dataType == "BYTE") return DATA_INT;
	else if(dataType == "EPOCH" || dataType == "ISO_TIME" || dataType == "EPOCH16") return DATA_TIME;
	else if( dataType == "ISO_TIME_RANGE") return DATA_EVENT;
	else return DATA_STR;
	
}

DvString QieData::DataTypeToStr(int dataType){

	if(dataType == DATA_DBL) return DvString("DOUBLE") ;
	else if(dataType == DATA_FLT) return DvString("FLOAT") ;
	else if(dataType == DATA_INT) return DvString("INT");
	else if(dataType == DATA_TIME) return DvString("ISO_TIME");
	else if( dataType == DATA_EVENT) return DvString("ISO_TIME_RANGE");
	else return DvString("CHAR");

}


// ------------------------------------------------

void QieData::countEntries(int &gattr_datatype_0, int &num_elems, vector <string> &valueList)
{
	int n_entries=0;
	gattr_datatype_0 = DATA_STR;
	int gattr_datatype = DATA_STR;

	size_t posStart;
	size_t posEnd;
   	// read the global block 
   	while(inFile.readLine(posStart, posEnd) != '\0')
	{
		if( !parseLine(inFile.content, posStart, posEnd) )  {
			string nextLine(&(inFile.content[posStart]), posEnd - posStart + 1);
		      QieMsg( "WARNING", "Failed to parse line in Global block =", nextLine.c_str());
			 // failed to parse line
			continue;
		}
       	if(parameter == "END_META") return;
       	
		if(parameter == "G_ATTR_DATA_TYPE" || parameter == "VALUE_TYPE"){
			
				// current data type
				gattr_datatype = getDataType(value.clean());  // deep copy
				
				// once first entry is found, don't change datatype (QSAS can't cope)
				if(n_entries == 0) gattr_datatype_0 = getDataType(value);  	
							
        	  	continue;
		}

		if(parameter == "ENTRY")
		{
			if(n_entries == 0) 
			{
				num_elems = countEntryElements(value, inFile.attr_delim);
			}
			if(gattr_datatype == gattr_datatype_0)
			{
				valueList.push_back(value.c_str());
				n_entries++;
			}
		}
	}
	
	return;
}

// ------------------------------------------------

bool QieData::readGlobal(const char *gAttrName)
{
	// make local copy as parameter/value re-used
	string globalAttrName(gAttrName);
	int gattr_datatype;
	int gattr_datatype_0;
	int num_elems_0;
	
	vector <string> valueList;
	
	// count number of entries
	countEntries(gattr_datatype_0, num_elems_0, valueList);
	
	if( globalAttrName != value.clean())  {
		// last line read should be end meta, so value contains meta name
		QieMsg( "WARNING", "Global block start/end names differ", globalAttrName.c_str(), value.c_str());
		return false;
	}

	int nEntries =  valueList.size();	
	gattr_datatype = gattr_datatype_0;
	
	if(nEntries == 0) {
		QieMsg( "ERROR", "no G entries counted");
		return false;
	}	
	if(num_elems_0 == 0) {
		QieMsg( "ERROR", "G entry has no elements");
		return false;
	}
	
    DvObject_var Gobj;

	if(gattr_datatype == DATA_STR  )
	{
		vector <size_t> dim(1);
		dim[0] = num_elems_0;
		Gobj = new DvObject(DvString(""), dim, nEntries);
		for(int n=0; n < nEntries; n++)
		{
			int numEntryElements = countEntryElements(valueList[n], inFile.attr_delim);
			if(num_elems_0 != numEntryElements)
			{	
				QieMsg( "WARNING", "Global entries number of elements differ");
				continue;
			}
			int startAt = 0;
			string token;
			for(int i=0; i<num_elems_0; i++){
				token = getNextEntry(valueList[n], startAt, inFile.attr_delim);
				Gobj->str(n * num_elems_0 + i).setQuoted(token);
			}
		}
	}
	else if(gattr_datatype == DATA_DBL || gattr_datatype == DATA_FLT  )
	{
		vector <size_t> dim(1);
		dim[0] = num_elems_0;
		Gobj = new DvObject(0.0, dim, nEntries);
		for(int n=0; n < nEntries; n++)
		{
			int numEntryElements = countEntryElements(valueList[n], inFile.attr_delim);
			if(num_elems_0 != numEntryElements)
			{	
				QieMsg( "WARNING", "Global entries number of elements differ");
				continue;
			}
			
			int startAt = 0;
			string token;
			for(int i=0; i<num_elems_0; i++){
				token = getNextEntry(valueList[n], startAt, inFile.attr_delim);
				Gobj->dbl(n * num_elems_0 + i) = strtod(token.c_str(), NULL);
			}
		}
	}
	else if( gattr_datatype == DATA_INT )
	{
		vector <size_t> dim(1);
		dim[0] = num_elems_0;
		Gobj = new DvObject((int) 0, dim, nEntries);
		for(int n=0; n < nEntries; n++)
		{
			int numEntryElements = countEntryElements(valueList[n], inFile.attr_delim);
			if(num_elems_0 != numEntryElements)
			{	
				QieMsg( "WARNING", "Global entries number of elements differ");
				continue;
			}
			
			int startAt = 0;
			string token;
			for(int i=0; i<num_elems_0; i++){
				token = getNextEntry(valueList[n], startAt, inFile.attr_delim);
				Gobj->itg(n * num_elems_0 + i) = (int) strtol(token.c_str(), NULL, 10);
			}
		}
	}
	else if(gattr_datatype == DATA_TIME){
		if(num_elems_0 == 1)
		{
			Gobj = new DvObject(DvTime(), nEntries);
			for(int n=0; n < nEntries; n++)
			{
				int numEntryElements = countEntryElements(valueList[n], inFile.attr_delim);
				if(1 != numEntryElements)
				{	
					QieMsg( "WARNING", "Global entries number of elements differ");
					continue;
				}
				Gobj->time(n).setFromISOstring(DvString(valueList[n]).cstr());
			}
			Gobj->secRes = getSecResolution(valueList[0]);
			
		}
		else if(num_elems_0 == 2)
		{
			Gobj = new DvObject(DvEvent(), nEntries);
			for(int n=0; n < nEntries; n++)
			{
				int numEntryElements = countEntryElements(valueList[n], inFile.attr_delim);
	
				if(num_elems_0 != numEntryElements)
				{	
					QieMsg( "WARNING", "Global entry for interval does not have two date/times");
					continue;
				}
	
				int startAt = 0;
				string token;
				token = getNextEntry(valueList[n], startAt, inFile.attr_delim);
				DvTime ts;
				ts.setFromISOstring(token.c_str());
	
				token = getNextEntry(valueList[n], startAt, inFile.attr_delim);
				DvTime te;
				te.setFromISOstring(token.c_str());
		
				Gobj->event(n).set(ts, te);
				if(n==0) Gobj->secRes = getSecResolution(token);
			}
		}
		else
		{
			vector <size_t> dim(1);
			dim[0] = num_elems_0;
			Gobj = new DvObject(DvTime(), dim, nEntries);
			for(int n=0; n < nEntries; n++)
			{
				int numEntryElements = countEntryElements(valueList[n], inFile.attr_delim);
				if(num_elems_0 != numEntryElements)
				{	
					QieMsg( "WARNING", "Global entries number of elements differ");
					continue;
				}
			
				int startAt = 0;
				string token;
				for(int i=0; i<num_elems_0; i++){
					token = getNextEntry(valueList[n], startAt, inFile.attr_delim);
					Gobj->time(n * num_elems_0 + i).setFromISOstring(token.c_str());
					if(i==0) Gobj->secRes = getSecResolution(token);
				}
			}
		}
	}
	else if(gattr_datatype == DATA_EVENT){
			
		vector <size_t> dim(1);
		dim[0] = num_elems_0;
		Gobj = new DvObject(DvEvent(), dim, nEntries);
		
		for(int n=0; n < nEntries; n++)
		{
			int numEntryElements = countEntryElements(valueList[n], inFile.attr_delim);
			if(num_elems_0 != numEntryElements)
			{	
				QieMsg( "WARNING", "Global entries number of elements differ");
				continue;
			}
			
			int startAt = 0;
			string token;
			for(int i=0; i<num_elems_0; i++){
				
				token = getNextEntry(valueList[n], startAt, inFile.attr_delim);

				string rangeStr = DvString(token).cstr(); // strips quotes
				int pos = rangeStr.find("/");
				string startStr = rangeStr.substr(0, pos);
				string endStr = rangeStr.substr(pos+1);
				DvTime ts;
				ts.setFromISOstring(startStr.c_str());
				DvTime te;
				te.setFromISOstring(endStr.c_str());
				
				Gobj->event(n * num_elems_0 + i).set(ts, te);
				if(i==0) Gobj->secRes = getSecResolution(startStr);
			}
		}
	}
	
	if( Gobj.is_nil() ) {
		return false;
	}
		  
	DvString typeXref = DataTypeToStr(gattr_datatype_0);
	Gobj->change_xref("value_type", typeXref);
	
	GlobalList.append(globalAttrName, Gobj);

	return true;
}

// ------------------------------------------------

QieVariable * QieData::initVariable(const char *variableName, QieFile *vFile)
{
	// initialise variable and format first (set defaults)
	
	if(vFile == NULL) vFile = &inFile;
	QieVariable *V = new QieVariable(variableName);
	
	V->n_posn = recordNumEntries;
	std::string dataValues=" ";
	int nEntries = 1;
	size_t posStart;
	size_t posEnd;
	while(vFile->readLine(posStart, posEnd) != '\0'){
		if( !parseLine(vFile->content, posStart, posEnd) ) {
			string nextLine(&(vFile->content[posStart]), posEnd - posStart + 1);
			QieMsg( "WARNING", "Failed to parse line in Variable block =", nextLine.c_str());
			continue;
		}

		if(parameter == "END_VARIABLE") {
        		if(V->name != value.clean()) {
		     	 	QieMsg( "WARNING", "Variable block start/end names differ", V->name.c_str(), value.c_str());
         		}
          		break; // end variable block, now make object
		}

		if(parameter == "DATA_TYPE" || parameter == "VALUE_TYPE"){
				V->datatype = getDataType(value.clean()); 
				// Note: first time var is default time tags
				if(d0_var_num == -1 && (value == "ISO_TIME" || value == "EPOCH")) d0_var_num = QieVars.size(); // not pushed back yet
				continue;
		}
		
		if(parameter == "DATA") {
			/* NOVARY variable */
			V->recVary = false;
			dataValues = value.c_str(); // deep copy for creating object after while loop completes
			continue;
		}
		
		if(parameter == "COLUMN_START") {
			V->col_start = atol(value.cstr());
			continue;
		}

		if(parameter == "COLUMN_WIDTH") {
			V->col_width = atol(value.cstr());
			V->entry_len = V->col_width;
			continue;
		}
		 
		if(parameter == "TIME_START") {
           	V->ft.ft_offset.setFromISOstring(value.clean()); 
			V->ft.use_ft_offset = true;
          	continue;
		}
		
		if(parameter == FILL_VALUE) {
           	if( value.clean() == "NaN") V->NaNstr = "NaN";
			// fall through and read
		}
		
		if(parameter == "ALL_RECORDS_FORMAT") {
           	V->ft.ft_all_time_fmt = value.clean(); 
          	continue;
		}
		if(parameter == "ALL_RECORDS_TIME") {
           	V->ft.ft_all_time = value.clean(); 
          	continue;
		}
		
		if(parameter == "YY_OFFSET") {
			V->ft.yy_offset = (int) strtol(value.cstr(), NULL, 10);
          	continue;
		}
		if(parameter == "MM_OFFSET") {
			V->ft.mm_offset = (int) strtol(value.cstr(), NULL, 10);
          	continue;
		}
		if(parameter == "DD_OFFSET") {
			V->ft.dd_offset = (int) strtol(value.cstr(), NULL, 10);
          	continue;
		}
		if(parameter == "SS_OFFSET") {
			V->ft.ss_offset = strtof(value.cstr(), NULL);
          	continue;
		}

		if(parameter == "TIME_FORMAT") {
			if(value.clean() == "FREE_TIME_FORMAT") {
   		       V->useFreeFormat = true;
			}
			else if(value.clean() == "ISO"){
     	       V->useFreeFormat = false;
			}         
			else{
				QieMsg( "WARNING", "Unknown time format designation '", value.cstr(), "' being ignored");
			}
			continue;
		}

		if(parameter == "TIME_FORMAT_STRING") {
			V->ft.ft_format = value.cstr();
			V->ft.parseFreeTime();
          	continue;
		}

		if(parameter == "MAXSTRLEN") {
			// no longer needed
			continue;
		}
		if(parameter == "SIZES") {

			int n_elems = countEntryElements(value, vFile->attr_delim);
			// set up Dim object 
			
			if(n_elems == 0)  continue;  
		
			V->objDim.clear();
			V->objDim.assign(n_elems, 0);	
			int startAt = 0;
			string token;
			for(int i=0; i<n_elems; i++){
					
				token = getNextEntry(value, startAt, vFile->attr_delim);
				int Ndim = (int) strtol(token.c_str() , NULL, 10);
				if(Ndim == 0) {
					QieMsg( "ERROR", "Variable dimension size zero from '", value.c_str());
					return NULL;
				}
				nEntries = nEntries * Ndim;
				V->objDim[i] = Ndim;
			}
			continue;
		}
        	 /* else read attribute */
		DvObject_var nextAttr = this->QieText_to_Attr(value, vFile); // V needed for type and dimensions

		if( nextAttr.is_ok() ){
			V->Vobj->change_xref(parameter.c_str(), nextAttr);
		}
	} // end while loop

	// If NRV object, fill with data now
	if(dataValues != " "){
        DvObject_var holdXrefs = V->Vobj;
		V->Vobj = QieText_to_Obj(V, dataValues, vFile);
        V->Vobj->copy_xrefs_from(holdXrefs);
	}

	if (V->recVary) recordNumEntries += nEntries;

	return V;  
}

// ----------------------------------------------------------

DvObject * QieData::QieText_to_Attr(DvString valueStr, QieFile *vFile)
{
	int datatype;
	DvObject * dobj;
	size_t posStart;
	size_t posEnd;
	
	if(vFile == NULL) vFile = &inFile;
	
	static DvString NaNstr("NaN");
	
	// concatenate strings written over several lines
	if(valueStr[0] == '\"'){
		size_t ptr = valueStr.rfind('\"'); // look for closing quote
		while(ptr == string::npos && vFile->readLine(posStart, posEnd) != '\0'){ // quote not closed
			valueStr += " ";
			valueStr += string(&(vFile->content[posStart]), posEnd - posStart +1 );
			ptr = valueStr.rfind('\"'); // keep looking for closing quote
		}
	}
	// concatenate strings written over several lines
	if(valueStr[0] == '{'){
		// Added for MAG Rosetta LBL files
		size_t ptr = valueStr.rfind('}'); // look for closing quote
		while(ptr == string::npos && vFile->readLine(posStart, posEnd) != '\0'){ // quote not closed
			valueStr += " ";
			valueStr += string(&(vFile->content[posStart]), posEnd - posStart +1 );
			ptr = valueStr.rfind('}'); // keep looking for closing quote
		}
		valueStr = valueStr.between('{', '}');
		valueStr = valueStr.replaceAll("\r", "");
		valueStr = valueStr.replaceAll("\n", "");
	}
	
	// break strings written over multiple lines 
	if(valueStr[0] == '\"'){
		DvString delim = "\" ";
		delim[1] = vFile->attr_delim;
		delim += "\"";
		valueStr = valueStr.replaceAll("\r", "");
		size_t ptr = valueStr.rfind("\n");
		if(ptr != string::npos){
			// remove last \n
			valueStr[ptr] = ' ';
		}
		valueStr = valueStr.replaceAll("\n", delim.c_str());
		// remove first entry if empty
		DvString first = valueStr.betweenIncl('\"', '\"');
		if(first.blank()) {
			DvString extra(vFile->attr_delim);
			first += extra;
			valueStr = valueStr.after(first.c_str());
		}
	}

    // Remove opening and closing brackets
    if(valueStr[0] == '(' && valueStr.endsIn(")") ){
        valueStr = valueStr.between('(', ')');
    }
    
	// create data object from valueStr
	int n_elems = countEntryElements(valueStr, vFile->attr_delim);
	if( n_elems == 0) return new DvObject();
   
	int startAt = 0;
	string token;
	token = getNextEntry(valueStr, startAt, vFile->attr_delim);

	if(token[0] == '\"') datatype = DATA_STR;
	else if( NaNstr == token.c_str() ) datatype = DATA_DBL;
	else if(QieIsISOtimeRange(token)) datatype = DATA_EVENT;
	else if(QieIsISOtime(token)) datatype = DATA_TIME;
	else{
       // test for numeric data type 
		char * endPtr;
		strtod(token.c_str(), &endPtr);
		if( token == string(endPtr) || *endPtr != '\0') datatype = DATA_STR; // not just numeric value found, use char as safe
		else if(token.find('.') != string::npos || token.find('e') != string::npos || token.find('E') != string::npos) datatype = DATA_DBL;
		else  datatype = DATA_INT;
	}
		
	if(datatype == DATA_STR)
	{
		vector <size_t> dim(1);
		dim[0] = n_elems;
		dobj = new DvObject(DvString(""), dim);

		dobj->str(0).setQuoted(token);
		for(int i=1; i<n_elems; i++){
			token = getNextEntry(valueStr, startAt, vFile->attr_delim);
			dobj->str(i).setQuoted(token);
		}
	}
	else if(datatype == DATA_TIME)
	{		
		if(n_elems == 1){
	
			dobj = new DvObject(DvTime(token)); 
			dobj->secRes = getSecResolution(token);
		}
		else if(n_elems == 2){
			// assume to be a time interval (NRV) 
			DvTime iv_start(token.c_str());
			token = getNextEntry(valueStr, startAt, vFile->attr_delim);
			DvTime iv_end(token.c_str());
			dobj = new DvObject(DvEvent(iv_start, iv_end)); 
			 
			dobj->secRes = getSecResolution(token);
		}
		else{ 
			// treat as time sequence
			dobj = new DvObject(DvTime(), n_elems);
			dobj->time(0).setFromISOstring(token.c_str());
			for (int i=1; i < n_elems; i++){
				token = getNextEntry(valueStr, startAt, vFile->attr_delim);
				dobj->time(i).setFromISOstring(token.c_str());
			}	
			dobj->secRes = getSecResolution(token);					
		}
	}
	else if(datatype == DATA_EVENT)
	{		
	
		dobj = new DvObject(DvEvent(), n_elems);
		dobj->event(0).setFromISOstring(token.c_str());
		for (int i=1; i < n_elems; i++){
			token = getNextEntry(valueStr, startAt, vFile->attr_delim);
			dobj->event(i).setFromISOstring(token.c_str());
		}		
		dobj->secRes = getSecResolution(token);					
			
	}
   	else if(datatype == DATA_INT )
	{
		vector <size_t> dim(1);
		dim[0] = n_elems;
		dobj = new DvObject((int) 0, dim);
		dobj->itg(0) =  (int) strtol(token.c_str(), NULL, 10);

		for(int i=1; i<n_elems; i++){
			token = getNextEntry(valueStr, startAt, vFile->attr_delim);
			dobj->itg(i) =  (int) strtol(token.c_str(), NULL, 10);
		}
	}
   	else if(datatype == DATA_DBL || datatype == DATA_FLT )
	{
		vector <size_t> dim(1);
		dim[0] = n_elems;
		dobj = new DvObject((double)0.0, dim);
		if(NaNstr == token.c_str() ) dobj->dbl(0) = DvNaN;
		else dobj->dbl(0) =  strtod(token.c_str(), NULL);
		for(int i=1; i<n_elems; i++){
			token = getNextEntry(valueStr, startAt, vFile->attr_delim);
			if(NaNstr == token.c_str()) dobj->dbl(i) = DvNaN;
			else dobj->dbl(i) =  strtod(token.c_str(), NULL);
		}
	}
	else {
		dobj = new DvObject();
	}
	
    return dobj;
}


// ----------------------------------------------------------

DvObject * QieData::QieText_to_Obj(QieVariable *V, std::string valueStr, QieFile *vFile)
{
	if(vFile == NULL) vFile = &inFile;
	
	// used for NRV data, uses attr_delim
	DvObject * dobj;
	
	// create data object from valueStr
	size_t num_elems = countEntryElements(valueStr, vFile->attr_delim);
	if(num_elems == 0) {
		dobj = new DvObject();
		return dobj;
	}
	if(V->datatype == DATA_STR)
	{
		vector <size_t> dim(1);
		dim[0] = num_elems;
		int startAt=0;
		string token;
		dobj = new DvObject(DvString(""), dim);
		for(size_t i=0; i<num_elems; i++){
			token = getNextEntry(valueStr, startAt, vFile->attr_delim);
			dobj->str(i).setQuoted(token);
		}
	}
	else if(V->datatype == DATA_EVENT)
	{				
		int startAt=0;
		string token;
		dobj = new DvObject(DvEvent(), num_elems);
		for(size_t i=0; i<num_elems; i++){
			token = getNextEntry(valueStr, startAt, vFile->attr_delim);
			dobj->event(i).setFromISOstring(token.c_str());
		}	
		dobj->secRes = getSecResolution(token);		
	}
	else if(V->datatype == DATA_TIME)
	{		
		if(num_elems == 1){
			int startAt=0;
			string token;
			token = getNextEntry(valueStr, startAt, vFile->attr_delim);
			dobj = new DvObject(DvTime(token.c_str() ));
			dobj->secRes = getSecResolution(token);
		}
		else if(num_elems == 2){
			// assume to be a time interval (NRV) 

			int startAt=0;
			string token;
			token = getNextEntry(valueStr, startAt, vFile->attr_delim);
			DvTime iv_start(token.c_str());
			token = getNextEntry(valueStr, startAt, vFile->attr_delim);
			DvTime iv_end(token.c_str());
			dobj= new DvObject(DvEvent(iv_start, iv_end)); 
			dobj->secRes = getSecResolution(token);
		}
		else{ 
			// treat as time sequence

			dobj = new DvObject(DvTime(), num_elems);
			int startAt=0;
			string token;
			for (size_t i=0; i < num_elems; i++){
				token = getNextEntry(valueStr, startAt, vFile->attr_delim);
				dobj->time(i).setFromISOstring(token.c_str());
			}
			dobj->secRes = getSecResolution(token);
		}
	}
   	else if( V->datatype == DATA_INT )
	{
		int startAt=0;
		string token;
		dobj = new DvObject((int) 0, V->objDim);
			
		if(num_elems != dobj->arraySize()) return dobj;

		for(size_t i=0; i<num_elems; i++){
			token = getNextEntry(valueStr, startAt, vFile->attr_delim);
			dobj->itg(i) = (int) strtol(token.c_str(), NULL, 10);
		}
	}
   	else if(V->datatype == DATA_DBL || V->datatype == DATA_FLT)
	{
		int startAt=0;
		string token;
		dobj = new DvObject(0.0, V->objDim);

		if(num_elems != dobj->arraySize()) return dobj;
			
		for(size_t i=0; i<num_elems; i++){
			token = getNextEntry(valueStr, startAt, vFile->attr_delim);
			dobj->dbl(i) = strtod(token.c_str(), NULL);
		}
	}
	else dobj = new DvObject();
	
	return dobj;

}


// ----------------------------------------------------------

void QieData::clearGetThis()
{
	//  sets the getThis flag to false for all vars 	
	for (size_t i=0; i< QieVars.size(); i++){
		QieVars[i]->getThis = false;
	}
}

// ----------------------------------------------------------

void QieData::setGetThis(int vnum)
{
	// recursively sets the getThis flag to true for var and xrefs
	QieVariable *V = QieVars[vnum];

	if(V->getThis) return; // already set
	
	V->getThis = true;
	 
	// go through xrefs and find vars pointed to
	DvNode *Xnode = V->Vobj->first_xref();

	while(Xnode){
		if(Xnode->obj().is_nil()) {
			Xnode = Xnode->next;
			continue;
		}
		// if ptr to var get var
		if( Xnode->obj()->is_str() ){
			int attr_num = findVarNumByName( Xnode->obj()->str().c_str());
			if(attr_num != -1) {
			
				// found dependent variable
				setGetThis(attr_num);
			}
		}
		Xnode = Xnode->next;
	}
}

// ----------------------------------------------------------

void QieData::initSequence(QieVariable * V, int nRecs)
{
    // copy dummy obj as it has xrefs
    DvObject_var holdXrefs = V->Vobj;
    
	// determine dimensionality
	V->nEntries = 1;
	for(size_t i=0; i<V->objDim.size(); i++){
		V->nEntries *= V->objDim[i];
	}
		
	// initialise sequence 
	if(V->datatype == DATA_DBL){
		if(V->nEntries == 1) V->Vobj = new DvObject(0.0, nRecs);	
		else  V->Vobj = new DvObject(0.0, V->objDim, nRecs);				
	}
	else if(V->datatype == DATA_FLT){
		if(V->nEntries == 1) V->Vobj = new DvObject(0.0, nRecs);	
		else  V->Vobj = new DvObject(0.0, V->objDim, nRecs);				
	}
	else if (V->datatype == DATA_TIME){

		// time sequence
		if(V->nEntries == 1) V->Vobj= new DvObject(DvTime(), nRecs);
		else V->Vobj= new DvObject(DvTime(), V->objDim, nRecs);
		
		if(V->useFreeFormat){
			// storage for accumulating time
			V->ft.year.assign(nRecs*V->nEntries, 0 );
			V->ft.month.assign(nRecs*V->nEntries, 0 );
			V->ft.day.assign(nRecs*V->nEntries, 0 );
			V->ft.dayOfYear.assign(nRecs*V->nEntries, 0 );
			V->ft.hour.assign(nRecs*V->nEntries, 0.0 );
			V->ft.minute.assign(nRecs*V->nEntries, 0.0 );
			V->ft.second.assign(nRecs*V->nEntries, 0.0 );
		}
	
	}
	else if(V->datatype == DATA_STR){
		if(V->nEntries == 1) V->Vobj = new DvObject(DvString(""), nRecs);
		else V->Vobj = new DvObject(DvString(""), V->objDim, nRecs);
	}
	else if (V->datatype == DATA_EVENT){

		// time interval sequence
		if(V->nEntries == 1) V->Vobj = new DvObject(DvEvent(), nRecs);	
		else 	V->Vobj = new DvObject(DvEvent(), V->objDim, nRecs);
	}
	else if (V->datatype == DATA_INT){

		// time interval sequence
		if(V->nEntries == 1) V->Vobj = new DvObject((int) 0, nRecs);	
		else 	 V->Vobj = new DvObject((int) 0, V->objDim, nRecs);
	}
	else{
		//  double 
		if(V->nEntries == 1) V->Vobj = new DvObject(0.0, nRecs);	
		else  V->Vobj = new DvObject(0.0, V->objDim, nRecs);
	}
	
	V->nRecs = nRecs;
    
    V->Vobj->copy_xrefs_from(holdXrefs);
    
	return;
}

// ----------------------------------------------------------

DvObject_var QieData::readTT_TAB(size_t vnum)
{
	// first ensure that all variable have been read and post processed

	vector <int> getList;

	// initialise variables
	for(size_t i=0; i<QieVars.size(); i++){

		if( QieVars[i]->recVary){
			// excludes NRV vars for flat files as read in header
			if(QieVars[i]->Vobj.is_ok()) {
				continue; 
			}
		
			initSequence(QieVars[i], nRecords);
			getList.push_back(i);			
		}
	}
	if(getList.size() > 0){
		// read records for all vars
		for( size_t recNow=0; recNow<nRecords; recNow++ ) 
		{
			nEntry = -1; // number of entry in record
			recVarEntry = inFile.recPosn[recNow]; // location of start of record
			for(size_t nv=0; nv<getList.size(); nv++){
				getVarRecordTableEntry(getList[nv], recNow);
			}
						
		}
		postRunRosetta();
	}
	
	// now get TT for requested var
	
	DvObject_var tt;
	QieVariable *var = QieVars[vnum];

	// find time tags variable for vnum
	DvNode *tnode = var->Vobj->find_xref(DEPEND_0);
	if(tnode)
	{
		tt = tnode->obj();
		if(tt.is_ok() && tt->is_time()) return tt; // case that tt object has been determined
		
		if(tt->is_str()) {
			int ttVarNum = findVarNumByName( tt->str().c_str());
			if(ttVarNum >= 0) tt = QieVars[ttVarNum]->Vobj;
		}
	}
	return tt;
	
}

// ----------------------------------------------------------

DvObject_var QieData::readTT(size_t vnum, size_t nRecs)
{
 	if(inFile.extn == ".TAB" ) return readTT_TAB(vnum);
	
	// Used only to get time vars, does not attach xrefs
	QieVariable *var = QieVars[vnum];
						
	DvObject_var tt;
	// find time tags variable
	DvNode *tnode = var->Vobj->find_xref(DEPEND_0);
	if(tnode)
	{
		tt = tnode->obj();
		if(tt.is_ok() && tt->is_time()) return tt; // case that tt object has been determined
		
		if(tt->is_str()) {
			int ttVarNum = findVarNumByName( tt->str().c_str());

			if(ttVarNum == -1) return new DvObject();

			// check if this has been read before
			tt = QieVars[ttVarNum]->Vobj;
			if(tt.is_nil() || tt->seqSize() != nRecs)
			{
				if(inFile.extn == ".CDF"){
					clearGetThis();
					setGetThis(ttVarNum); 

					// create list of vars to fetch and initialise
					vector <size_t> getList;
					for(size_t i=0; i<QieVars.size(); i++){
						if(QieVars[i]->getThis)  getList.push_back(i);
					}
					
					// Now we have Themis components too, get everything
					getCDFvarDataList(getList, 0, -1);
				}
				else if(inFile.extn == ".QFT" )
				{
					// first call for this variable
					initSequence(QieVars[ttVarNum], nRecs);
					
					for( size_t recNow=0; recNow<nRecs; recNow++ ) 
					{
						nEntry = -1; // number of entry in record
						recVarEntry = inFile.recPosn[recNow]; // location of start of record
						getVarRecordTableEntry(ttVarNum, recNow);
						
					}


				}
				else if(inFile.data_delim == ' ')
				{
					// special delimited type as multiple spaces allowed as single delim
					// first call for this variable
					initSequence(QieVars[ttVarNum], nRecs);
					
					for( size_t recNow=0; recNow<nRecs; recNow++ ) 
					{
						nEntry = -1; // number of entry in record
						recVarEntry = inFile.recPosn[recNow]; // location of start of record
						getVarRecordEntrySpSep(ttVarNum, recNow);
						
					}

				}
				else 
				{
					// first call for this variable
					initSequence(QieVars[ttVarNum], nRecs);

					for( size_t recNow=0; recNow<nRecs; recNow++ ) 
					{
						nEntry = -1; // number of entry in record
						recVarEntry = inFile.recPosn[recNow]; // location of start of record
						getVarRecordEntry(ttVarNum, recNow);
						
					}

				}
                
                // post process Free Time variables to assemble complete time
                if(QieVars[ttVarNum]->useFreeFormat ) QieVars[ttVarNum]->processTime();

                QieVars[ttVarNum]->ensureDeltas();
                
                tt = QieVars[ttVarNum]->Vobj;
			}
			// else tt already read, will use them
		}
	}
	else{
		// variable has no DEPEND_0, can we make event table into time sequence?
		if( var->Vobj->is_event() ){
			int len = var->Vobj->seqSize();
			tt = new DvObject(DvTime(), len);
			for(int i=0; i<len; i++){
				tt->time(i) = var->Vobj->event(i).start();
			}
		}
		
	}
	
 

	return tt;
}
// ----------------------------------------------------------

bool QieData::readVarData(int vnum)
{
	clearGetThis(); // ensure no vars are set to be read
	if( QieVars.size() <= 0 ) {
		QieMsg ("ERROR:", "No variables detected in header");
		return false;
	}
	size_t nRec = nRecords; // nRecords may get changed for Rosetta (.TAB) files
	size_t recStart = 0;
	size_t recEnd =  nRec - 1 ;
	
	QieVariable *V = QieVars[vnum];
	// NOTE: NRV variables are read already in queryFile() but not queryCDFfile()
	if ( !(V->recVary) && inFile.extn != ".CDF") return true; 

	// recursively set the getThis flag to true for var and variable type xrefs (incl time tags)
	setGetThis(vnum); 

	// create list of vars to fetch and initialise
	vector <size_t> getList;
	for(size_t i=0; i<QieVars.size(); i++){
		// force Rosetta TAB files to get data for ALL vars as they may need to fix record lengths
		if(inFile.extn == ".TAB") QieVars[i]->getThis = true; 
		
		if(QieVars[i]->getThis){

			if(inFile.extn == ".CDF"){
			
			 	getList.push_back(i);
			}
			else if( QieVars[i]->recVary){
				// excludes NRV vars for flat files as read in header
				if(QieVars[i]->Vobj.is_ok()) {
					// variable already read
					QieVars[i]->getThis = false;
					continue; 
				}
				initSequence(QieVars[i], nRec);
				getList.push_back(i);

			}
		}
	}

	if(getList.size() == 0) return true; // done already
	

	if(inFile.extn == ".CDF") {
	   // fetch data for variables from CDF file

	   getCDFvarDataList(getList, recStart, recEnd);
	}
	else if (inFile.extn == ".QFT" ){

		// get records requested for tabular data
		size_t recNum=0;
		
		for( size_t recNow=recStart; recNow<=recEnd; recNow++ ){
			// locate start and length of each token in record
		
			nEntry = -1; // number of entry in record
			recVarEntry = inFile.recPosn[recNow]; // location of start of record
			for(size_t nv=0; nv<getList.size(); nv++){
				getVarRecordTableEntry(getList[nv], recNum);
			}

			recNum++;
		
		} // end while over records

	    // The variables needed have now been read into their Vobj pointers for the records requested
	}
	else if (inFile.extn == ".TAB"){

		// get records requested for tabular data
		size_t recNum=0;
		
		for( size_t recNow=recStart; recNow<=recEnd; recNow++ ){
			// locate start and length of each token in record
		
			nEntry = -1; // number of entry in record
			recVarEntry = inFile.recPosn[recNow]; // location of start of record
			for(size_t nv=0; nv<getList.size(); nv++){
				getVarRecordTableEntry(getList[nv], recNum);
			}

			recNum++;
		
		} // end while over records

	    // The variables needed have now been read into their Vobj pointers for the records requested
		postRunRosetta();

		
	}
	else if(inFile.data_delim == ' '){
		// get records requested for appropriate data type
		size_t recNum=0;
		
		for( size_t recNow=recStart; recNow<=recEnd; recNow++ ){
			// locate start and length of each token in record
		
			nEntry = -1; // number of entry in record
			recVarEntry = inFile.recPosn[recNow]; // location of start of record
			for(size_t nv=0; nv<getList.size(); nv++){
				getVarRecordEntrySpSep(getList[nv], recNum);
			}

			recNum++;
		
		} // end while over records

	    // The variables needed have now been read into their Vobj pointers for the records requested
		postRunRosetta();

	}
	else{
		// get records requested for appropriate data type
		size_t recNum=0;

		for( size_t recNow=recStart; recNow<=recEnd; recNow++ ){
			// locate start and length of each token in record
		
			nEntry = -1; // number of entry in record
			recVarEntry = inFile.recPosn[recNow]; // location of start of record
			for(size_t nv=0; nv<getList.size(); nv++){
				getVarRecordEntry(getList[nv], recNum);
			}

			recNum++;
		
		} // end while over records

	    // The variables needed have now been read into their Vobj pointers for the records requested
	}
    
	// post process Free Time variables to assemble complete time
	for(size_t i=0; i<QieVars.size(); i++){
		if(QieVars[i]->getThis){
			QieVariable *V = QieVars[i];	
			if(V->datatype == DATA_TIME ){

				if(V->useFreeFormat){
					// first partial time var is the one we attach
					if(mergeTimeVar(i)){
						V->processTime();
						break; // only one broken up time possible
					}
					V->processTime();
				}
			}
            V->ensureDeltas();
		}
	}

	return true;
}		

// -----------------------------------------------------

bool QieData::mergeTimeVar(int vnum){
	// recursively adds up elements from partial times spread over several variables
	QieVariable *V = QieVars[vnum];

	DvNode* node = V->Vobj->find_xref("Time_Component");

	if(node && node->obj().is_ok() && node->obj()->is_str()) {
		int nNext = findVarNumByName(node->obj()->str().c_str());

		if(nNext >= 0){
			mergeTimeVar(nNext); // recursive
			
			if(V->nRecs*V->nEntries == QieVars[nNext]->nRecs*QieVars[nNext]->nEntries){
				QieVars[nNext]->skip = true; // don't import or write as now merged
				// seconds field is likely to be in sub-component, check sec resolution
				if(V->ft.NdecSec < QieVars[nNext]->ft.NdecSec) V->ft.NdecSec = QieVars[nNext]->ft.NdecSec;

				if(QieVars[nNext]->ft.use_DoY) V->ft.use_DoY = true;
				for(size_t i=0; i<V->nRecs*V->nEntries; i++){
					V->ft.year[i] += QieVars[nNext]->ft.year[i];
					V->ft.month[i] += QieVars[nNext]->ft.month[i];
					V->ft.day[i] += QieVars[nNext]->ft.day[i];
					V->ft.dayOfYear[i] += QieVars[nNext]->ft.dayOfYear[i];
					V->ft.hour[i] += QieVars[nNext]->ft.hour[i];
					V->ft.minute[i] += QieVars[nNext]->ft.minute[i];
					V->ft.second[i] += QieVars[nNext]->ft.second[i];

				}
			}
			return true;
		}
	}
	return false; // not partial var
}


// ------------------------------------------------

void QieData::attachXrefs(int varN, DvString varName){

	
    // remove Themis epoch xrefs as tt now cleaned up
    QieVars[varN]->Vobj->delete_xref("VIRTUAL");
    QieVars[varN]->Vobj->delete_xref("COMPONENT_0");
    QieVars[varN]->Vobj->delete_xref("COMPONENT_1");
    
    
    // Change attributes that point to vars. 
	
	DvNode *Xnode = QieVars[varN]->Vobj->first_xref();
	
	while(Xnode){

		// Find attributes that point to vars
		if( Xnode->obj()->is_str() ){
			// attach var pointed to as xref if this is a var name
			int attr_num = findVarNumByName( Xnode->obj()->str().c_str());

			if(attr_num != -1	&& Xnode->obj()->str() != QieVars[varN]->name )// skip self reference, e.g. Ptr_var_name
			{ 
				// use var as attribute, 
				QieVars[varN]->Vobj->change_xref(Xnode->name(), QieVars[attr_num]->Vobj);
				
				DvString ptrVarName(QieVars[attr_num]->name.c_str());
				QieVars[attr_num]->Vobj->change_xref(PTR_VAR_NAME, ptrVarName);
				
                // recursively attach xrefs
				attachXrefs(attr_num, QieVars[varN]->name);
				Xnode = Xnode->next;
				continue;
				
			}
		}
		
		// otherwise attach attribute as it is
		QieVars[varN]->Vobj->change_xref(Xnode->name(), Xnode->obj());
			
		Xnode = Xnode->next;
	}

	// Ensure some common SI_Conversions from units
	if( !QieVars[varN]->Vobj->xref_exists(SI_CONVERSION)){
		if( QieVars[varN]->Vobj->xref_exists(UNITS)){
			DvObject_var units = QieVars[varN]->Vobj->get_xref(UNITS);
			if(units.is_ok()) {
				// add to this list for known incomplete files
				if(units->asStr() == "s") QieVars[varN]->Vobj->change_xref(SI_CONVERSION, "1>s");
				else if(units->asStr() == "ms") QieVars[varN]->Vobj->change_xref(SI_CONVERSION, "1.e-3>s");
				else if(units->asStr() == "us") QieVars[varN]->Vobj->change_xref(SI_CONVERSION, "1.e-6>s");
				else if(units->asStr() == "ns") QieVars[varN]->Vobj->change_xref(SI_CONVERSION, "1.e-9>s");
			}
		}
	}
	
	// If time type ensure deltas are in seconds
	if(QieVars[varN]->Vobj->is_time()){
		DvObject_var dp = QieVars[varN]->Vobj->get_xref(DELTA_PLUS);
		if(dp.is_ok()) {
			dp = dp->changeUnitsTo("1>s", "s");
			QieVars[varN]->Vobj->change_xref(DELTA_PLUS, dp);
		}
		DvObject_var dm = QieVars[varN]->Vobj->get_xref(DELTA_MINUS);
		if(dm.is_ok()) {
			dm = dm->changeUnitsTo("1>s", "s");
			QieVars[varN]->Vobj->change_xref(DELTA_MINUS, dm);
		}
	}
}

// -----------------------------------------------
char * QieData::getNextRecEntry(){

	while(*recVarEntry == '\n'|| *recVarEntry == '\r') recVarEntry++;

	while(*recVarEntry == '!') {
		recVarEntry++;
		while(*recVarEntry != '\n') recVarEntry++; // skip rest of line
		recVarEntry++; // now skip '\n'
	}
	
	char * entry = recVarEntry;
	
	// find next delimiter or end of line 
	while(*recVarEntry != inFile.end_of_line 
		&& *recVarEntry != inFile.data_delim  
		&& *recVarEntry != '\0') recVarEntry++;
	
//	*recVarEntry = '\0'; // replace delimiter with null
	if(*recVarEntry!= '\0') recVarEntry++; // skip over delimiter
	nEntry++;

	return entry;
	
}

// -----------------------------------------------
string QieData::getNextRecEntryStr(){

	while(*recVarEntry == '\n'|| *recVarEntry == '\r') recVarEntry++;
	size_t len = 0;
	
	char * entry = recVarEntry;
	
	// find next delimiter or end of line 
	while(*recVarEntry != inFile.end_of_line 
		&& *recVarEntry != inFile.data_delim  
		&& *recVarEntry != '\0') {
			recVarEntry++;
			len ++;
	}
	
//	*recVarEntry = '\0'; // replace delimiter with null
	recVarEntry++; // skip over delimiter
	nEntry++;
	
	return string(entry, len);
	
}

// -----------------------------------------------
void QieData::setRecEntry(int n){
	// queue up so next getNextRecEntry() gets entry number n
	while(nEntry < n-1) {
		// find next delimiter or end of line 
		while( *recVarEntry != inFile.data_delim && *recVarEntry != '\0') {
			recVarEntry++;
		}
		recVarEntry++; // skip delimiter
		nEntry++;
	}

	while(*recVarEntry == ' ') recVarEntry++; // remove leading space
	
	return;
	
}

// ------------------------------------------------

void QieData::getVarRecordEntry(int nToGet, int recNum){

	// recNum is position in output sequence
	// Note variables are in same order as data in record	
	
	QieVariable *V = QieVars[nToGet];
	char *entry = NULL;
	
	setRecEntry(V->n_posn);
	int nEntries = V->nEntries;
	int arrOffset = recNum * nEntries;

	if(V->datatype == DATA_DBL){
		// double
		if(V->NaNstr.empty()){
			for( int i=arrOffset; i<arrOffset + nEntries; i++){
				entry = getNextRecEntry();
				V->Vobj->dbl(i) = strtod(entry, NULL);
			}
		}
		else{
			for( int i=arrOffset; i<arrOffset + nEntries; i++){
				entry = getNextRecEntry();
				if('n' == entry[0] || 'N' == entry[0] ) V->Vobj->dbl(i) = DvNaN;
				else V->Vobj->dbl(i) = strtod(entry, NULL);
			}
		}
	}
	else if(V->datatype == DATA_FLT){
		// double
		
		for( int i=arrOffset; i<arrOffset + nEntries; i++){
			entry = getNextRecEntry();
			V->Vobj->dbl(i) = strtod(entry, NULL);
		}

	}
	else if(V->datatype == DATA_TIME){
		if(V->useFreeFormat){
			for(int i=0; i<nEntries; i++){
				entry = getNextRecEntry();
				V->ft.assembleTime(arrOffset + i, entry);
			}
		}
		else{
			for(int i=0; i<nEntries; i++){
				V->Vobj->time(arrOffset + i).setFromISOstring(getNextRecEntryStr());
			}
		}
	}
	else if(V->datatype == DATA_STR)
	{
		for(int i=0; i<nEntries; i++){
			V->Vobj->str(arrOffset + i) = getNextRecEntryStr();
			
		}
	}
	else if(V->datatype == DATA_EVENT){
	
		for(int i=0; i<nEntries; i++){
			V->Vobj->event(arrOffset + i).setFromISOstring(getNextRecEntryStr());
		}
	}
	else if(V->datatype == DATA_INT){
	
		for(int i=0; i<nEntries; i++){
			entry = getNextRecEntry();
			V->Vobj->itg(arrOffset + i) = (int) strtol(entry, NULL, 10);
		}
	}
	return;
}		

// -----------------------------------------------
char * QieData::getNextRecEntrySpSep(){

	while(*recVarEntry == '\n'|| *recVarEntry == '\r') recVarEntry++;
	
	char * entry = recVarEntry;
	
	// find next delimiter or end of line 
	while(*recVarEntry != inFile.end_of_line 
		&& *recVarEntry != ' '  
		&& *recVarEntry != '\0') recVarEntry++;
	
	*recVarEntry = '\0'; // replace delimiter with null
	recVarEntry++; // skip space
	
	while( *recVarEntry == ' '  ) recVarEntry++; // remove extra spaces
	
	nEntry++;
	return entry;
	
}

// -----------------------------------------------
string QieData::getNextRecEntryStrSpSep(){

	while(*recVarEntry == '\n'|| *recVarEntry == '\r') recVarEntry++;

	size_t len=0;
		
	char * entry = recVarEntry;
	
	// find next delimiter or end of line 
	while(*recVarEntry != inFile.end_of_line 
		&& *recVarEntry != ' '  
		&& *recVarEntry != '\0') {
			len++;
			recVarEntry++;
	}
//	*recVarEntry = '\0'; // replace delimiter with null
	recVarEntry++; // skip this delimiter
	
	while( *recVarEntry == ' '  ) recVarEntry++; // remove extra spaces
	
	nEntry++;
	
	return string(entry, len);
	
}

// -----------------------------------------------
void QieData::setRecEntrySpSep(int n){
	
	// queue up so next getNextRecEntry() gets entry number n
	while(nEntry < n-1) {

		// find next delimiter or end of NULL from earlier var read 
		while( *recVarEntry != ' ' &&  *recVarEntry != '\0' ) recVarEntry++;
		
		while( *recVarEntry == ' '  || *recVarEntry == '\0' ) recVarEntry++; // remove extra spaces
		
		nEntry++;
	}
	return;
	
}

// ------------------------------------------------

void QieData::getVarRecordEntrySpSep(int nToGet, int recNum){

	// recNum is position in output sequence
	// Note variables are in same order as data in record	
	
	
	QieVariable *V = QieVars[nToGet];
	char *entry = NULL;
	
	setRecEntrySpSep(V->n_posn);
	
	int nEntries = V->nEntries;
	int arrOffset = recNum * nEntries;

	if(V->datatype == DATA_DBL){
		// double
		if(V->NaNstr.empty()){
			for( int i=arrOffset; i<arrOffset + nEntries; i++){
				entry = getNextRecEntrySpSep();
				V->Vobj->dbl(i) = strtod(entry, NULL);
			}
		}
		else{
			for( int i=arrOffset; i<arrOffset + nEntries; i++){
				entry = getNextRecEntrySpSep();
				if('n' == entry[0] || 'N' == entry[0] ) V->Vobj->dbl(i) = strtod(entry, NULL); 
				else V->Vobj->dbl(i) = DvNaN;
			}
		}

	}
	else if(V->datatype == DATA_FLT){
		// double
		
		for( int i=arrOffset; i<arrOffset + nEntries; i++){
			entry = getNextRecEntrySpSep();
			V->Vobj->dbl(i) = strtod(entry, NULL);
		}

	}
	else if(V->datatype == DATA_TIME){
		if(V->useFreeFormat){
			for(int i=0; i<nEntries; i++){
				entry = getNextRecEntrySpSep();			
				V->ft.assembleTime(arrOffset + i, entry);
			}
		}
		else{
			for(int i=0; i<nEntries; i++){
				V->Vobj->time(arrOffset + i).setFromISOstring(getNextRecEntryStrSpSep());
			}
		}
	}
	else if(V->datatype == DATA_STR)
	{
		for(int i=0; i<nEntries; i++){
			V->Vobj->str(arrOffset + i) = getNextRecEntryStrSpSep();
		}
	}
	else if(V->datatype == DATA_EVENT){
	
		for(int i=0; i<nEntries; i++){
			V->Vobj->event(arrOffset + i).setFromISOstring(getNextRecEntryStrSpSep());
		}
	}
	else if(V->datatype == DATA_INT){
	
		for(int i=0; i<nEntries; i++){
			entry = getNextRecEntrySpSep();
			V->Vobj->itg(arrOffset + i) = (int) strtol(entry, NULL, 10);
		}
	}
	return;
}		

// ------------------------------------------------

void QieData::getVarRecordTableEntry(int nToGet, int recNum){

	// recNum is position in output sequence
	// Note variables are in same order as data in record	
	

	QieVariable *V = QieVars[nToGet];
	string entry;
	int nEntries = V->nEntries;
	int arrOffset = recNum * nEntries;

	if(V->datatype == DATA_DBL){
		// double

		if(V->NaNstr.empty()){
			for( int i=0; i<nEntries; i++){
				entry = string(recVarEntry  + V->col_start + i*V->col_width, V->entry_len);
				V->Vobj->dbl(arrOffset + i) = strtod(entry.c_str(), NULL); 
			}
		}
		else{
			for( int i=0; i<nEntries; i++){
				entry = string(recVarEntry  + V->col_start + i*V->col_width, V->entry_len);
				if('n' == entry[0] || 'N' == entry[0] ) V->Vobj->dbl(arrOffset + i) = strtod(entry.c_str(), NULL); 
				else V->Vobj->dbl(arrOffset + i) = DvNaN;
			}
		}


	}
	else if(V->datatype == DATA_FLT){
		// double
		
		for( int i=0; i<nEntries; i++){
			entry = string(recVarEntry  + V->col_start + i*V->col_width, V->entry_len);
			V->Vobj->dbl(arrOffset + i) = strtod(entry.c_str(), NULL);
		}

	}
	else if(V->datatype == DATA_TIME){
		if(V->useFreeFormat){
			for(int i=0; i<nEntries; i++){
				entry = string(recVarEntry  + V->col_start + i*V->col_width, V->entry_len);		
				V->ft.assembleTime(arrOffset + i, entry.c_str());
			}
            
		}
		else{
			for(int i=0; i<nEntries; i++){
				entry = string(recVarEntry  + V->col_start + i*V->col_width, V->entry_len);
				V->Vobj->time(arrOffset + i).setFromISOstring(entry);
			}
		}
	}
	else if(V->datatype == DATA_STR)
	{
		for(int i=0; i<nEntries; i++){
			entry = string(recVarEntry  + V->col_start + i*V->col_width, V->entry_len);
			V->Vobj->str(arrOffset + i) = DvString(entry);
		}
	}
	else if(V->datatype == DATA_EVENT){
	
		for(int i=0; i<nEntries; i++){
			entry = string(recVarEntry  + V->col_start + i*V->col_width, V->entry_len);
			V->Vobj->event(arrOffset + i).setFromISOstring(entry);
		}
	}
	else if(V->datatype == DATA_INT){
	
		for(int i=0; i<nEntries; i++){
			entry = string(recVarEntry  + V->col_start + i*V->col_width, V->entry_len);
			V->Vobj->itg(arrOffset + i) = (int) strtol(entry.c_str(), NULL, 10);
		}
	}
	return;
}		


// ------------------------------------------------

bool QieData::parseLine(char *nextLine, size_t posStart, size_t posEnd)
{  
  size_t ptr;
  // posnEnd is posn of end marker, so don't go there

  for( ptr=posStart; ptr<posEnd; ptr++) 
  		if(nextLine[ptr] == '=') break ;
	
  if(ptr >= posEnd) {
  	string strLine(&(nextLine[posStart]), posEnd - posStart + 1);
	QieMsg ("NOTE", "Missing '=', Not metadata line:- \n", strLine.c_str());
    return false;
  }
  if(ptr == 0) {
  	string strLine(&(nextLine[posStart]), posEnd - posStart + 1);
	QieMsg ("ERROR", "Line starts '=', cannot parse line\n", strLine.c_str());
    return false;
  }

  /* remove leading white space in param */
  while( (nextLine[posStart] == ' ' || nextLine[posStart] == '\t') ) posStart++;
 
   /* remove possible trailing white space in param */
  size_t tokLen = ptr-1; 
  while( nextLine[tokLen] == ' ' || nextLine[tokLen] == '\t' ) tokLen--;
  tokLen++;
  tokLen -= posStart;
  if(tokLen <= 0) {
  	string strLine(&(nextLine[posStart]), posEnd - posStart );
	QieMsg ("ERROR", "No parameter, cannot parse line\n", strLine.c_str());
    return false;
  }
  parameter = string( &(nextLine[posStart]), tokLen);
 
  // clean quotes off parameter, but value may need quotes
  parameter.clean();
  
  ptr ++;
  // remove leading white space in value 
  while( nextLine[ptr] == ' ' || nextLine[ptr] == '\n' || nextLine[ptr] == '\t' )  ptr++;
 
  // remove newline and possible trailing white space in value 
  
  posEnd--; // lose line end marker (usually \n)
  
  while( nextLine[posEnd] == ' ' || nextLine[posEnd] == '\t' ) posEnd--;
  posEnd++;
  if(posEnd <= ptr) {
	// replace empty RHS with single space
 cout << " empty " << endl;fflush(stdout);
	value = string(" ");
  }
  else{
	value = string( &(nextLine[ptr]), posEnd-ptr);
  }

  return true;

} 




// -----------------------------------------------

size_t QieData::countEntryElements(std::string &inputStr, const char delim)
{
	int ptr = 0;
	int LineLen = inputStr.length();
	const char *cStr = inputStr.c_str();
	if(LineLen == 0) return 0;
	
	int numElems = 1;
	
	// start on entry content
	while (cStr[ptr] == ' ') ptr++;
	while(ptr < LineLen) {

		if(cStr[ptr] == delim)
		{
			numElems++;
			// find next entry content (needed for space delim)
			while (cStr[ptr] == ' ') ptr++;
			
		}
		else if(cStr[ptr] == '\"') {
			// ignore anything INSIDE quotes
			ptr++;
			while(cStr[ptr] != '\"' && ptr < LineLen) ptr++;
		}
		ptr++;
	}
	return numElems;
}
	
// -----------------------------------------------

string QieData::getNextEntry(string &inputStr, int &startAt, const char delim){
		
	int ptr = startAt;
	int LineLen = inputStr.length();
	const char *cStr = inputStr.c_str();
	int truncateAt;
	string token;

	// don't start looking until actual entry found
	while(cStr[ptr] == ' ') ptr++;
	
	
	// peel off entries

	while(ptr < LineLen )
	{
		// find next delim if there is one
		if( cStr[ptr] == delim) break;

		// skip to end of quoted text, ignore delim INSIDE quotes
		if(cStr[ptr] == '\"') {
			ptr++;

			while(cStr[ptr] != '\"' ) 
			{
				ptr++;
				if(ptr == LineLen) 
				{
					// back up as it gets incremented again
					ptr--;
					break;
				}
			}
		}
		ptr++;
	}
	// ptr is now just passed end of line or on next delimiter
	// remove leading white space 

	while( (cStr[startAt] == ' ' || cStr[startAt] == '\t')  && startAt < ptr) startAt++;

	// remove trailing white space and quotes
	truncateAt = ptr-1;
	while ((cStr[truncateAt] == ' ' || cStr[truncateAt] == '\n' ) && startAt < truncateAt) truncateAt--;

	if( truncateAt - startAt >= 0 ) 	
		token = inputStr.substr(startAt, truncateAt - startAt +1);
	else
		token = " ";
		
	ptr++;
	startAt = ptr;

	return token;
}


// -----------------------------------------------

int QieData::findVarNumByName(const char * name){

	for(size_t i=0; i<QieVars.size(); i++){
		if(QieVars[i]->name == DvString(name) ) return i;
	}
	
	return -1; // not found
}


// --------------------------------------------------------

DvObject_var QieData::getD0_obj(int vnum){

	DvObject_var Nil_obj;
	int tt_num = d0_var_num; // pick up file default if set
	
	DvNode *Xnode = QieVars[vnum]->Vobj->find_xref(DEPEND_0);
	if(Xnode){
		DvObject_var D0 = Xnode->obj();
				
		if( D0.is_ok() && D0->is_str()){
			tt_num = findVarNumByName(D0->str(0).c_str());
		}
		else return D0; // object itself not pointer to obj
		
		if (tt_num == -1) tt_num = d0_var_num; // failed, so reset to default
	}
	if (tt_num != -1) {
		return QieVars[tt_num]->Vobj;
	}

	return Nil_obj;
}

// --------------------------------------------------------

void QieData::ensureDependZero(){

	// ensure an attribute called DEPEND_0 exists if time tags found
    if (d0_var_num == -1) return; // no tt variable
    
	for(size_t i=0; i< QieVars.size(); i++){
		// NRV variables should not need DEPEND_0 adding
		if( !(QieVars[i]->recVary) ) continue;
        
		// if any variable has DEPEND_0 set leave it alone
		if(QieVars[i]->Vobj->find_xref(DEPEND_0)) continue;
        
        // Half interval and other time tag attrs don't take depend_0
        DvString SIC = QieVars[i]->Vobj->getXrefText(SI_CONVERSION);
        if(SIC.after('>') == "s") continue;

        // object must be same length as Dep0
        if(QieVars[i]->Vobj->seqSize() != QieVars[d0_var_num]->Vobj->seqSize()) continue;
        
		if (d0_var_num != (int)i){
			// set default tt. Note variable may not be read yet so create ptr to var attribute
			DvString tName =  QieVars[d0_var_num]->name;
			DvObject_var xref = new DvObject(tName);
			QieVars[i]->Vobj->change_xref(DEPEND_0, xref);
		}
	}
	return;

}


// -----------------------------------------------------------------------------------------
//    Rosetta TAB/LBL File type  
//    -------------------

// ----------------------------------------------------------

int QieData::queryTABfile(DvString fullName)
{
	// TAB files are postProcessed to clean up Rosetta PDS syntax
	// This requires all variables to be read when file opened
	DvString LBLfile = fullName;
	LBLfile = LBLfile.replace(".TAB", ".LBL");
	LBLfile = LBLfile.replace(".tab", ".lbl");

	inFile.openFile(LBLfile.c_str());
	
	if(inFile.ifp.is_open()){
	
		nRecords = read_LBL();
		
		inFile.closeFile();
		
		getAttributeHelper(fullName.c_str());

		inFile.openFile(fullName);
		
	}
	else return 0;
	

	size_t nFound = inFile.countRecords(); // locate start of each record
	if(nFound != nRecords) {
		QieMsg ("WARNING", "Number of records in .TAB file differs from .LBL FILE_RECORDS entry\n");
		cout << " n Records found =" << nFound << ", expected = " << nRecords << " " << fullName << endl;
		nRecords = nFound;
	}

	// for TAB file read all variables NOW as Rosetta data spread over records
	// this fixes nRecords for import
	readVarData(0);

	return nRecords;
}

// ----------------------------------------------------------

int QieData::read_LBL()
{
	int numRecords=0;
	DvString delim="";
	
 	// CAA defaults may not be written in the file, so set in advance
	inFile.attr_delim = ',';
	inFile.data_delim = ' '; // not needed

	// note: readLine skips comment lines 
	size_t posStart;
	size_t posEnd;
	
	while(inFile.readLine(posStart, posEnd) != '\0'){
		if( !parseLine(inFile.content, posStart, posEnd) ){
			// skip lines not of form parameter = value
			continue;
		}
		
		if(parameter == "FILE_RECORDS" ) {
				numRecords = (int) strtol(value.cstr(), NULL, 10);
			continue;
		}
		
		if(parameter == "DELIMITER" ) {
			if(value.c_str()[0] != ' ') delim.setQuoted(value);
			continue;
		}
	  
	  
		if(parameter == "OBJECT") {
			if( value.clean() == "COLUMN" ) initColumn(delim.size());
			else if( value == "HEADER" ) inFile.hasTitle=true;
			continue;
		}
		
		if(parameter == "COLUMNS" ||
		   parameter == "ROWS" ||
		   parameter == "END" ||
		   parameter == "END_OBJECT" ||
		   parameter == "^TABLE") {
			// skip some Rosetta stuff
			continue;
		}

		// else global
		DvObject_var Gobj = this->QieText_to_Attr(value); // V needed for type and dimensions

		if( Gobj.is_ok() ){
			GlobalList.append(parameter, Gobj);
		}


	} // end while

	// ensure there is a DEPEND_0 set if time line exists
	ensureDependZero();

	return numRecords;
} 

// ------------------------------------------------

bool QieData::initColumn(int delimLen)
{
	// initialise variable for Rosetta
	QieVariable *V = new QieVariable();
	
	V->n_posn = recordNumEntries;

	std::string dataValues=" ";
	int nEntries = 1;
	size_t posStart;
	size_t posEnd;
	bool haveItemOffset=false;
	bool haveItemBytes=false;
	int itemOffset=0;
	int itemBytes=0;
	
	while(inFile.readLine(posStart, posEnd) != '\0'){
		if( !parseLine(inFile.content, posStart, posEnd) ) {
			continue;
		}
		if(parameter == "NAME") {
			V->name.setQuoted(value);
			continue;
		}
		
		if(parameter == "END_OBJECT") {
			break; // end variable block, now make object
		}
		
		if(parameter == "START_BYTE") {
			V->col_start = atol(value.cstr())-1; // Rosetta counts from 1
			continue;
		}

		if(parameter == "BYTES") {
			V->col_width = atol(value.cstr())+delimLen; 
			V->entry_len = V->col_width;
			continue;
		}
		
		if(parameter == "ITEMS") {
			nEntries = atol(value.cstr());
			
			V->objDim.clear();
			V->objDim.assign(1, nEntries);	// for now since we don't have dims
			continue;
		}
		
		if(parameter == "MISSING_CONSTANT") {
			DvObject_var nextAttr = this->QieText_to_Attr(value); 

			if( nextAttr.is_ok() ){
				V->Vobj->change_xref(FILL_VALUE, nextAttr);
			}

			continue;
		}
		
		if(parameter == "ITEM_BYTES") {
			itemBytes = atol(value.cstr());
			haveItemBytes=true;
			continue;
		}
		
		if(parameter == "ITEM_OFFSET") {
			itemOffset = atol(value.cstr());
			haveItemOffset=true;
			continue;
		}
		
		
		if(parameter == "VALID_MAXIMUM") {
			DvObject_var nextAttr = this->QieText_to_Attr(value); // V needed for type and dimensions

			if( nextAttr.is_ok() ){
				V->Vobj->change_xref("VALIDMAX", nextAttr);
			}
			continue;
		}
		
		if(parameter == "VALID_MINIMUM") {
			DvObject_var nextAttr = this->QieText_to_Attr(value); // V needed for type and dimensions

			if( nextAttr.is_ok() ){
				V->Vobj->change_xref("VALIDMIN", nextAttr);
			}
			continue;
		}

		if(parameter == "DATA_TYPE" ){
		
			V->datatype = getLBLDataType(value.clean()); 
			
			// Note: first time var is default time tags
			if(d0_var_num == -1 && (value == "TIME" || value == "EPOCH")) d0_var_num = QieVars.size(); // not pushed back yet
			continue;
		}

		if(parameter == "TIME_FORMAT") {
			if(value.clean() == "FREE_TIME_FORMAT") {
   		       V->useFreeFormat = true;
			}
			else if(value == "ISO"){
     	       V->useFreeFormat = false;
			}         
			else{
                V->useFreeFormat = false;
                QieMsg( "WARNING", "Unknown time format designation '", value.c_str(), "' being ignored");
			}
			continue;
		}

		if(parameter == "TIME_FORMAT_STRING") {
			V->ft.ft_format = value.cstr();
			V->ft.parseFreeTime();
          	continue;
		}
		 
		/* else read attribute */
		DvObject_var nextAttr = this->QieText_to_Attr(value); // V needed for type and dimensions

		if( nextAttr.is_ok() ){
			V->Vobj->change_xref(parameter.c_str(), nextAttr);
		}
		
	} // end while loop

	// If ITEM_OFFSET is provided we don't need DELIMITER
	// If ITEM_BYTES is provided BYTES is total for all entries in variable
	// DELIMITER is not needed if ITEM_OFFSET is present
	
	if(haveItemOffset) V->col_width = itemOffset;
	if(haveItemBytes) V->entry_len = itemBytes;
	
	
	QieVars.push_back(V);
	if (V->recVary) recordNumEntries += nEntries;

	return true;  
}

// ------------------------------------------------

int QieData::getLBLDataType(DvString dataType)
{
	if(dataType == "ASCII_REAL" ) return DATA_DBL;
	else if(dataType == "ASCII_INTEGER" ) return DATA_INT;
	else if( dataType == "TIME") return DATA_TIME;
	else if( dataType == "TIME_RANGE") return DATA_EVENT;
	else return DATA_STR;
	
}
// ------------------------------------------------

void QieData::postRunRosetta()
{
	// check for inst specific flag
	int repeat_4=1; // used for ICA dep4
	int PromFlag = -1;
    int timeIdx = -1;
	int promSection = 6;
	int iData = -1;
	
	// V6 ICA files and DEP_4 need explicit correction
	for(size_t i=0; i< QieVars.size(); i++){
		if(QieVars[i]->name == "PROM_SECTION") PromFlag = (int)i;
		if(QieVars[i]->name == "TIME_UTC") timeIdx = (int)i;
	}

	for(size_t i=0; i< QieVars.size(); i++){
		
		DvNode *Xnode = QieVars[i]->Vobj->find_xref("Post_Run_Qtran");
		if(Xnode){

			iData = i;
			
			if(	PromFlag != -1 && QieVars[PromFlag]->Vobj.is_ok() ) promSection = QieVars[PromFlag]->Vobj->asInt()-1;

			DvString Prun = Xnode->obj()->asStr();
			DvString inst = Prun.before(">");
			if(inst == "ICA"){

				if(promSection == 12 ){
					// table V6 set in .qat file
					repeat_4 = 1;
				}
				else if(promSection == 14 ){
					// table V4
					DvNode *XnodeDep4 = QieVars[i]->Vobj->find_xref("Depend_4");
					if(XnodeDep4){
						DvObject_var xref4 = new DvObject("Energy_Table_V4");
						XnodeDep4->setObj(xref4);
					}
				}
				else if(promSection == 11 || promSection == 8){
					// table V7
					repeat_4 = 3;
				//	QieVars[i]->Vobj->Dims()[0] /= repeat_4; // [0] because dimensions not yet accumulated
				//	QieVars[i]->objDim[0] /= repeat_4;
					DvNode *XnodeDep4 = QieVars[i]->Vobj->find_xref("Depend_4");
					if(XnodeDep4){
						DvObject_var xref4 = new DvObject("Energy_Table_V7");
						XnodeDep4->setObj(xref4);
					}
				}
				else if(promSection == 10){
					// table V8
					repeat_4 = 12;
				//	QieVars[i]->Vobj->Dims()[0] /= repeat_4; // [0] because dimensions not yet accumulated
				//	QieVars[i]->objDim[0] /= repeat_4;
					DvNode *XnodeDep4 = QieVars[i]->Vobj->find_xref("Depend_4");
					if(XnodeDep4){
						DvObject_var xref4 = new DvObject("Energy_Table_V8");
						XnodeDep4->setObj(xref4);
					}
				}
				else{
					// table V2
					repeat_4 = 1;
					DvNode *XnodeDep4 = QieVars[i]->Vobj->find_xref("Depend_4");
					if(XnodeDep4){
						DvObject_var xref4 = new DvObject("Energy_Table_V2");
						XnodeDep4->setObj(xref4);
					}
				}
			}
			QieVars[i]->Vobj->delete_xref("Post_Run_Qtran");
		}
	}

	// process free time
	for(size_t i=0; i< QieVars.size(); i++){
		if(QieVars[i]->useFreeFormat){
			QieVars[i]->processTime();
			QieVars[i]->useFreeFormat = false;
		}
	}

	// Merge component vars into vectors or arrays

	vector <DvString> toDelete;
	// combine array components
	for(size_t i=0; i< QieVars.size(); i++){
				
		DvString arrBase = "ARRAY_COMP_";
		int arrayLen = 1;
		DvString arrComp = arrBase + arrayLen;
		DvNode *Xnode = QieVars[i]->Vobj->find_xref(arrComp);

		// count array dimension
		while( Xnode ){
			arrayLen++;
			arrComp = arrBase + arrayLen;
			Xnode = QieVars[i]->Vobj->find_xref(arrComp);
		}
		arrayLen--; // last value was not found
		if(arrayLen < 1) continue;
		
		vector <size_t> dimObj;
		dimObj.push_back(arrayLen);
		DvObject_var vect = new DvObject( 0.0, dimObj, QieVars[i]->Vobj->seqSize());
		
		// get array elements in turn	
		for(size_t k=1; k<=(size_t)arrayLen; k++){
				
			arrComp = arrBase + (int)k;
			Xnode = QieVars[i]->Vobj->find_xref(arrComp);
			
			if(!Xnode || Xnode->obj().is_nil()) continue;
			
			int icomp = i;
			DvString compVar = Xnode->obj()->asStr();
			if(compVar != "THIS") {
				icomp = findVarNumByName(compVar.c_str());
                
				if(icomp == -1) continue;
				toDelete.push_back(QieVars[icomp]->name);
			}
			
			if(  QieVars[icomp]->Vobj.is_nil() ) continue;
			if( !QieVars[icomp]->Vobj->is_dbl() ) continue;
			
			// fill in component
			for(size_t n=0; n<QieVars[i]->Vobj->seqSize(); n++){
				vect->dbl(arrayLen*n + k - 1) = QieVars[icomp]->Vobj->dbl(n);
			}
			QieVars[i]->Vobj->delete_xref(arrComp);
		
		}
		
        DvObject_var holdXrefs = QieVars[i]->Vobj;
		QieVars[i]->Vobj=vect;
        QieVars[i]->Vobj->copy_xrefs_from(holdXrefs);
        
		DvNode *XnodeName = QieVars[i]->Vobj->find_xref("ARRAY_NAME");
		if(XnodeName) QieVars[i]->name = XnodeName->obj()->asStr();
		QieVars[i]->objDim = dimObj;
		QieVars[i]->nEntries = arrayLen;
	}

	// combine vector components
	for(size_t i=0; i< QieVars.size(); i++){
		
		DvNode *Xnode = QieVars[i]->Vobj->find_xref("VECTOR_COMP_1");

		if( Xnode ){
			DvNode *Xnode2 = QieVars[i]->Vobj->find_xref("VECTOR_COMP_2");
			DvNode *Xnode3 = QieVars[i]->Vobj->find_xref("VECTOR_COMP_3");
			if(!Xnode2 || !Xnode3) continue;
			
			int i1 = i;
			int i2 = i;
			int i3 = i;
			DvString comp1Var = Xnode->obj()->asStr();
			DvString comp2Var = Xnode2->obj()->asStr();
			DvString comp3Var = Xnode3->obj()->asStr();

			if(comp1Var != "THIS") i1 = findVarNumByName(comp1Var.c_str());
			if(comp2Var != "THIS") i2 = findVarNumByName(comp2Var.c_str());
			if(comp3Var != "THIS") i3 = findVarNumByName(comp3Var.c_str());

			if(i1 == -1 || i2 == -1 || i3 == -1) continue;
			
			if(QieVars[i1]->Vobj.is_nil() || QieVars[i2]->Vobj.is_nil() || QieVars[i3]->Vobj.is_nil() ) continue;
			
			vector <size_t> dimObj;
			dimObj.push_back(3);
			DvObject_var vect = new DvObject( 0.0, dimObj, QieVars[i]->Vobj->seqSize());
			for(size_t n=0; n<QieVars[i]->Vobj->seqSize(); n++){
				vect->dbl(3*n) = QieVars[i1]->Vobj->asDouble(n);
				vect->dbl(3*n+1) = QieVars[i2]->Vobj->asDouble(n);
				vect->dbl(3*n+2) = QieVars[i3]->Vobj->asDouble(n);
			}
            DvObject_var holdXrefs = QieVars[i]->Vobj;
			QieVars[i]->Vobj=vect;
            QieVars[i]->Vobj->copy_xrefs_from(holdXrefs);
			
            DvNode *XnodeName = QieVars[i]->Vobj->find_xref("VECTOR_NAME");
			QieVars[i]->name = XnodeName->obj()->asStr();
			QieVars[i]->objDim = dimObj;
			QieVars[i]->nEntries = 3;
			QieVars[i]->Vobj->delete_xref("VECTOR_COMP_1");
			QieVars[i]->Vobj->delete_xref("VECTOR_COMP_2");
			QieVars[i]->Vobj->delete_xref("VECTOR_COMP_3");
						
			toDelete.push_back(QieVars[i2]->name);
			toDelete.push_back(QieVars[i3]->name);
		}
	}
	
	// delete components
	for(size_t k=0; k<toDelete.size(); k++){
		int nDel = findVarNumByName(toDelete[k].c_str());
		if(nDel >= 0){
			delete QieVars[nDel];
			QieVars.erase(QieVars.begin()+nDel);
		}
	}

	toDelete.clear();

	// unpack dimensions spread over records
	for(int j=4; j>=1; j--){

		size_t length = 1;
		size_t nRecBefore = nRecords;
		
		DvString dep = "Depend_";
		dep += j;
		DvString alias = "DEP_";
		alias += j;
		alias += "_ALIAS";
		DvString aliasEnd = alias;
		aliasEnd += "_END";
		DvString fixed = "Fixed_Sizes_";
		fixed += j;
		
		DvMask msk(nRecords); // initialises to true (i.e. keep)
		bool useMask=false;
		
		for(size_t i=0; i< QieVars.size(); i++){
		
			// NRV variables should not need fixing
			if( !(QieVars[i]->recVary) ) continue;

			// if any variable has DEP_i_ALIAS convert repeat records to dimension
			DvNode *Xnode = QieVars[i]->Vobj->find_xref(alias);
			DvNode *XnodeEnd = QieVars[i]->Vobj->find_xref(aliasEnd);
			DvNode *SizesNode = QieVars[i]->Vobj->find_xref(fixed);
			
			// non-aliased, but special case
			if(j==1 && repeat_4 >1 && (int)i == iData){
					// energy (j==4) varies after mass and aximuth, but before elevation!
					// So do before elevation (j==1)
					// ICA energy bins repeated
					// add together for quicklook

				nRecords /= repeat_4;
				msk.resize(nRecords);
				
				QieVars[i]->Vobj = QieVars[i]->Vobj->aveRecEvery(repeat_4);
					
				// fix other vars
				for(size_t ii=0; ii< QieVars.size(); ii++){
					if(i==ii || !(QieVars[ii]->recVary) ) continue;
						
					QieVars[ii]->Vobj = QieVars[ii]->Vobj->sliceRecEvery(repeat_4);
				}
				nRecBefore /= repeat_4;
									
				QieVars[i]->nRecs = nRecords;
			}

			DvNode *XnodeDep = QieVars[i]->Vobj->find_xref(dep);
			if(XnodeDep){
				// Depend_i xref exists, add extra metadata
				DvString unitInfo = dep+"_Units";
				DvNode *XnodeDepUnits = QieVars[i]->Vobj->find_xref(unitInfo);
				if(XnodeDepUnits){
					XnodeDep->obj()->change_xref(UNITS, XnodeDepUnits->obj());
					QieVars[i]->Vobj->delete_xref(unitInfo);
				}

				DvString SICInfo = dep+"_SIC";
				DvNode *XnodeDepSIC = QieVars[i]->Vobj->find_xref(SICInfo);
				if(XnodeDepSIC){
					XnodeDep->obj()->change_xref(SI_CONVERSION, XnodeDepSIC->obj());
					QieVars[i]->Vobj->delete_xref(SICInfo);
				}
				
				DvString NameInfo = dep+"_Name";
				DvNode *XnodeDepName = QieVars[i]->Vobj->find_xref(NameInfo);
				if(XnodeDepName){
					XnodeDep->obj()->change_xref(FIELDNAM, XnodeDepName->obj());
					QieVars[i]->Vobj->delete_xref(NameInfo);
				}
				
				DvString StypeInfo = dep+"_SCALETYP";
				DvNode *XnodeDepStype = QieVars[i]->Vobj->find_xref(StypeInfo);
				if(XnodeDepStype){
					XnodeDep->obj()->change_xref(SCALETYP, XnodeDepStype->obj());
					QieVars[i]->Vobj->delete_xref(StypeInfo);
				}

			}
			
					
			if (Xnode){
				// Fixed array size specified, dimension may need spreading out
				Xnode->setName(dep);
				int depi = findVarNumByName(Xnode->obj()->asStr().c_str());

				if(SizesNode){
					DvNode *XnodeFill = QieVars[i]->Vobj->find_xref(FILL_VALUE);
					int fill = -1;
					if(XnodeFill && XnodeFill->obj().is_ok()){
						fill = XnodeFill->obj()->asInt();
					}
					
					useMask= true;
					if(SizesNode->obj().is_nil() || !SizesNode->obj()->is_int()) continue; // give up
					length = SizesNode->obj()->itg(0);

					int depiEnd = -1;
					if(XnodeEnd) {
						depiEnd = findVarNumByName(XnodeEnd->obj()->asStr().c_str());
						toDelete.push_back(XnodeEnd->obj()->asStr());
					}
					if(depi < 0 || QieVars[depi]->Vobj.is_nil() ) continue; // give up
					int arrLen = QieVars[i]->Vobj->arraySize();
					
					if(QieVars[depi]->Vobj->is_int()){
					
						nRecords = 0;

						// count records for repeating values (resolution in file changes)
						size_t startAt = 0;
						
						// find number of entries provided
						while(startAt < QieVars[depi]->Vobj->totalSize()){
							size_t entryLen=1;
							for(size_t k=1; k<length; k++ ){
								if( startAt + k >= QieVars[depi]->Vobj->totalSize() || QieVars[depi]->Vobj->asInt(startAt + k -1) >= QieVars[depi]->Vobj->asInt(startAt + k) ) break;
								entryLen++;
							}
							nRecords++;
							startAt += entryLen;
						}

						// fix dimensions, move up 1 place and insert length at bottom
						vector <size_t> newDim;
						newDim.push_back(length);
						for(size_t jj=0; jj<QieVars[i]->Vobj->dims.size(); jj++){
							newDim.push_back(QieVars[i]->Vobj->dims[jj]);
						}		
						
						// create new new object that is large enough 
						DvObject_var newObj = new DvObject((double)0.0, newDim, nRecords);
						
						
						// process repeating values
						startAt = 0;
						size_t ptr = 0;
						size_t rec = 0;
						
						while(startAt < QieVars[depi]->Vobj->totalSize()){
							size_t entryLen=1;
							for(size_t k=1; k<length; k++ ){
								if( startAt + k >= QieVars[depi]->Vobj->totalSize() || QieVars[depi]->Vobj->asInt(startAt + k -1) >= QieVars[depi]->Vobj->asInt(startAt + k) ) break;
								entryLen++;
							}
							
							// spread out data over this dim
							for(size_t p=0; p<entryLen; p++){
								// mark records for deletion in other vars (e.g. earlier depends or Dep_0)
								if(p>0) msk[startAt + p] = false;

								size_t column = (size_t) QieVars[depi]->Vobj->asInt(startAt + p);
								size_t columnEnd = column;
								if(depiEnd >= 0) columnEnd = QieVars[depiEnd]->Vobj->asInt(startAt + p);
								double nCols = columnEnd - column + 1;

								if(column < length) {
									for(size_t q=0; q<(size_t)arrLen; q++){
										for(size_t c=column; c<=columnEnd; c++){
											// repeat over summed bins
											if(QieVars[i]->Vobj->asInt(ptr) == fill) newObj->dbl(rec, arrLen*c + q) = (double) fill;
											else newObj->dbl(rec, arrLen*c + q) = QieVars[i]->Vobj->asDouble(ptr) / nCols;
										}
										ptr++;
									}
								}
							}
							
							rec++;
							startAt += entryLen;
						}

						// replace object
                        DvObject_var holdXrefs = QieVars[i]->Vobj;
                        QieVars[i]->Vobj = newObj;
                        QieVars[i]->Vobj->copy_xrefs_from(holdXrefs);						
					
						QieVars[i]->datatype = DATA_FLT; // average requires floating point type
						
						// replace depend_i
						valarray <int> newDepVal(length);
						for (size_t r=0; r<length; r++) newDepVal[r] = r;
						DvObject_var newDep = new DvObject(newDepVal);
						
						// replace index with bins if known
						DvNode *XnodeBins = QieVars[depi]->Vobj->find_xref("BINS");
						if(XnodeBins){
							DvString binsStr = XnodeBins->obj()->asStr(0);
							valarray <double> newBins(newDep->totalSize());
							newBins[0] = binsStr.toDouble();
							for(size_t k=1; k<newBins.size(); k++){
								binsStr = binsStr.after(',');
								newBins[k] = binsStr.toDouble();
							}
							
							newDep->replaceData(newBins);
							QieVars[depi]->Vobj->delete_xref("BINS");
							QieVars[depi]->Vobj->delete_xref("UNIT");
						}
				
						// convert to array
						newDep->seqLen = 1;
						newDep->dims.clear();
						newDep->dims.push_back(length);
						// replace variable
                        DvObject_var depXrefs = QieVars[depi]->Vobj;
						QieVars[depi]->Vobj = newDep;
                        QieVars[depi]->Vobj->copy_xrefs_from(depXrefs);
                        
						QieVars[depi]->recVary = false;
						QieVars[depi]->Vobj->delete_xref(DEPEND_0);
						QieVars[depi]->nRecs = 1;
					}
			
					QieVars[i]->nRecs = nRecords;
					QieVars[i]->objDim.clear();
					for(size_t k=0; k<QieVars[i]->Vobj->dims.size(); k++) QieVars[i]->objDim.push_back(QieVars[i]->Vobj->dims[k]);

					QieVars[i]->Vobj->delete_xref(fixed);
					QieVars[i]->Vobj->delete_xref(aliasEnd);

				}
				else{
					// no DEPEND_i_FIXED so dimension assumed same as 1st rec
					if(Xnode->obj().is_ok() && Xnode->obj()->is_str()){
						length = 1; // reset from earlier dim
					
						if(depi < 0 || QieVars[depi]->Vobj.is_nil() ) continue; // give up
						// count how many elements before repeat
						for(size_t k=1; k<QieVars[depi]->Vobj->totalSize(); k++ ){
							if( QieVars[depi]->Vobj->asInt(k-1) >= QieVars[depi]->Vobj->asInt(k) ) break;
							length++;
						}
						// reduce to single cycle
						DvObject_var newDep = QieVars[depi]->Vobj->subSequence(0, length-1);
						
						DvNode *XnodeBins = QieVars[depi]->Vobj->find_xref("BINS");
						if(XnodeBins){
							DvString binsStr = XnodeBins->obj()->asStr(0);
							valarray <double> newBins(newDep->totalSize());
							newBins[0] = binsStr.toDouble();
							for(size_t k=1; k<newBins.size(); k++){
								binsStr = binsStr.after(',');
								newBins[k] = binsStr.toDouble();
							}
							
							newDep->replaceData(newBins);
							QieVars[depi]->Vobj->delete_xref("BINS");
							QieVars[depi]->Vobj->delete_xref("UNIT");
						}

						
						// convert to array
						newDep->seqLen = 1;
						newDep->dims.clear();
						newDep->dims.push_back(length);
						// replace variable
                        DvObject_var depXrefs = QieVars[depi]->Vobj;
						QieVars[depi]->Vobj = newDep;
                        QieVars[depi]->Vobj->copy_xrefs_from(depXrefs);
 
						QieVars[depi]->recVary = false;
						QieVars[depi]->Vobj->delete_xref(DEPEND_0);
						QieVars[depi]->nRecs = 1;
					}
					else{
						length = Xnode->obj()->arraySize();
					}
				
				
					// fix dimensions, move up 1 place and insert length at bottom
					QieVars[i]->Vobj->dims.push_back(1);
					for(size_t j=QieVars[i]->Vobj->dims.size(); j>1; j--){
						QieVars[i]->Vobj->dims[j-1] = QieVars[i]->Vobj->dims[j-2];
					}
					QieVars[i]->Vobj->dims[0] = length;
					nRecords /= length;
					QieVars[i]->Vobj->seqLen = nRecords;
					QieVars[i]->nRecs = nRecords;
					QieVars[i]->objDim.clear();
					for(size_t k=0; k<QieVars[i]->Vobj->dims.size(); k++) QieVars[i]->objDim.push_back(QieVars[i]->Vobj->dims[k]);
				}
			}
			
		}

		if(length > 1){
			if(useMask){

			// use mask
				for(size_t i=0; i< QieVars.size(); i++){
					if( !(QieVars[i]->recVary) ) continue;
					// change nRecs for all vars so sliders in QDI work
					QieVars[i]->nRecs = nRecords;
					if( QieVars[i]->Vobj.is_ok() && QieVars[i]->Vobj->seqLen == nRecBefore) {
						QieVars[i]->Vobj->apply_mask(msk);
						QieVars[i]->nRecs = nRecords;
					}
				}
			}
			else{
				// subset other vars on time to match (all assumed to either have same extra dimension or have dim 1
				slice sl(0, nRecords, length);
				for(size_t i=0; i< QieVars.size(); i++){
					// change nRecs for all vars so sliders in QDI work
					QieVars[i]->nRecs = nRecords;
	
					if( QieVars[i]->Vobj.is_ok() && QieVars[i]->Vobj->seqLen == nRecBefore) {
						QieVars[i]->Vobj = QieVars[i]->Vobj->subSequence(sl);
						QieVars[i]->nRecs = nRecords;
					}
				}
			}
		}

	}

	for(size_t i=0; i< QieVars.size(); i++){

		// convert dimension index to index

		for (int k=1; k<5; k++){
			DvString depIdx = "Depend_";
			depIdx += k;
			depIdx += "_INDEX";

			DvNode *Xnode = QieVars[i]->Vobj->find_xref(depIdx);

			if( Xnode && Xnode->obj()->is_str()){
				int depidx_i = findVarNumByName(Xnode->obj()->str().c_str());
				
				DvString dep_str = "Depend_";
				dep_str += k;

				DvNode *XnodeDep = QieVars[i]->Vobj->find_xref(dep_str);
				if( XnodeDep && XnodeDep->obj()->is_str()){
					int dep_i = findVarNumByName(XnodeDep->obj()->str().c_str());

					size_t len = QieVars[depidx_i]->Vobj->seqSize(); // from index var
					vector <size_t> dimObj;
					size_t arrLen = QieVars[dep_i]->Vobj->Dims()[1]; // from dep var
					dimObj.push_back(arrLen);
					
					// create new t-dependent dep
					DvObject_var dep_var = new DvObject(0.0, dimObj, len);
					for(size_t n=0; n<len; n++){
						for(size_t j=0; j<arrLen; j++){
							dep_var->dbl(n, j) = QieVars[dep_i]->Vobj->asDouble(QieVars[depidx_i]->Vobj->asInt(n)*arrLen + j);
						}
					}
                    DvObject_var depXref = QieVars[dep_i]->Vobj;
					QieVars[dep_i]->Vobj = dep_var;
                    QieVars[dep_i]->Vobj->copy_xrefs_from(depXref);
				}
			}
		}
		// Create deltas for times if missing (LAP, MIP)
		if(QieVars[i]->name.contains("TIME") && (QieVars[i]->Vobj->is_time() || QieVars[i]->Vobj->is_dbl()) ){
		
			DvNode *XnodePlus = QieVars[i]->Vobj->find_xref("Delta_plus");
			DvNode *XnodeMinus = QieVars[i]->Vobj->find_xref("Delta_minus");
			if(XnodePlus && XnodeMinus) continue; // both exist already
		
// This is no use for data aliased to time as stop time is just end of that azimuth				
/*			DvNode *XnodeEndTime = QieVars[i]->Vobj->find_xref("END_TIME");
			int et_i = -1;
			if( XnodeEndTime ) et_i = findVarNumByName(XnodeEndTime->obj()->str().c_str());
			
			if(et_i > -1 && QieVars[et_i]->Vobj->seqSize() == QieVars[i]->Vobj->seqSize() ) {

				DvObject_var deltaObj = new DvObject((double)0.0, QieVars[i]->Vobj->seqSize());
				for(int n=0; n<QieVars[i]->Vobj->seqSize(); n++){
					deltaObj->dbl(n) = QieVars[et_i]->Vobj->asDouble(n) - QieVars[i]->Vobj->asDouble(n);
				}
*/				
			if(QieVars[i]->name.contains("START_TIME") 	){
				// Tag will be start time
				double delta = QieVars[i]->Vobj->get_spacing();
				DvObject_var deltaObj = new DvObject(delta);
				DvObject_var zeroObj = new DvObject(0.0);
				deltaObj->change_xref(SI_CONVERSION, "1>s");
				zeroObj->change_xref(SI_CONVERSION, "1>s");
				QieVars[i]->Vobj->change_xref("Delta_plus", deltaObj);
				QieVars[i]->Vobj->change_xref("Delta_minus", zeroObj);

			}
			else if(QieVars[i]->name.contains("STOP_TIME")) {
				// do nothing
			}
			else{
				QieVars[i]->ensureDeltas();
			}
		}

	}	

	// delete components
	for(size_t k=0; k<toDelete.size(); k++){
		int nDel = findVarNumByName(toDelete[k].c_str());
		if(nDel >= 0){
			delete QieVars[nDel];
			QieVars.erase(QieVars.begin()+nDel);
		}
	}

	return;

}

/**********************************************************************
************************ START OF CDF DATA EXPORT FUNCTIONS ***********
***********************************************************************/


bool QieData::writeFlat(bool replace, bool detached )
{
	int oldRes=3;
	// make sure time is first variable for CAA

	vector <QieVariable*> VoutOrder;
	vector <QieVariable*> VrecOrder;
	if(d0_var_num >= 0) {
		VoutOrder.push_back(QieVars[d0_var_num]);
		VrecOrder.push_back(QieVars[d0_var_num]);
		// set time output resolution
		if(dataSecRes >= 0) 
			oldRes = VoutOrder[0]->Vobj->time(0).setSecResolution(dataSecRes);
		else 
			oldRes = VoutOrder[0]->Vobj->time(0).setSecResolution(VoutOrder[0]->Vobj->secRes);

	}
	for(int n=0; n < (int) QieVars.size(); n++){
		if( n == d0_var_num) continue; // done already
		if(QieVars[n]->Vobj.is_ok() &&  !QieVars[n]->skip) VoutOrder.push_back(QieVars[n]);
		if(QieVars[n]->Vobj.is_ok() && QieVars[n]->recVary &&  !QieVars[n]->skip) VrecOrder.push_back(QieVars[n]);
	}
	QieVars = VoutOrder;
	d0_var_num = 0; // now 1st var
	
	// Write Header First
	if(detached )
	{
		if( ensureFileName(replace, detached) == false) {
			cerr << "Write Error: unable to open header file for write (file exists or no write permission)\n   " << outFile.name << ".qfh" << endl;
			return false;
		}
		
		writeHeader(detached);
		
		outFile.ofp.close();
		
		// check file name is safe to write to and open it
		if( ensureFileName(replace) == false) {
			cerr << "Write Error: unable to open file for write (file exists or no write permission)\n   " << outFile.name << outFile.extn <<endl;
			return false;
		}
		
	}
	else 
	{
		// check file name is safe to write to and open it
		if( ensureFileName(replace) == false) {
			cerr << "Write Error: unable to open file for write (file exists or no write permission)\n   " << outFile.name << outFile.extn << endl;
			return false;
		}
		
		writeHeader(detached); 
		// else NO_HEADER

	}

	
	// add data records
	QieVars = VrecOrder; // only write rec varying vars to records

	if(outFile.extn == ".qft") writeDataQFT(); 
	else if(outFile.extn == ".cef") writeDataCAA();
	else writeData();
	
	outFile.ofp.flush(); // flush buffer
			
	// reset time library
	if(d0_var_num >= 0) VoutOrder[0]->Vobj->time(0).setSecResolution(oldRes);
	
	return true;
}

// -------------------------------------------

void QieData::writeHeader(bool detached)
{
	if(outFile.extn == ".qft"){
		// magic number 
		outFile.ofp << "!@" << outFile.extn << "@" << DvString(QIE_VERSION_NUM) << endl;
		outFile.ofp << "!------------------- QSAS ASCII File -------------------|" << endl;
		outFile.ofp << "! Fixed column width table padded with CSDS fill values |" << endl;
		outFile.ofp << "! Column position in record counted from 0              |" << endl;
		outFile.ofp << "!                                                       |" << endl;
		outFile.ofp << "! ASCII Format                                          |" << endl;
		outFile.ofp << "! Native C ordering, last index varies fastest          |" << endl;
		outFile.ofp << "! Blank lines are ignored                               |" << endl;
		outFile.ofp << "! \"!\" escapes rest of line as comment                   |" << endl;
		outFile.ofp << "!-------------------------------------------------------|" << endl;
		outFile.ofp << "File_type = t" << endl;
    }
	else if(outFile.extn == ".qfd"){

		// magic number 
		outFile.ofp << "!@" << outFile.extn << "@" <<  DvString(QIE_VERSION_NUM) << endl;
		outFile.ofp << "!------------------- QSAS ASCII File -------------------|" << endl;
		outFile.ofp << "! Free width data entries separated by delimiters       |" << endl;
		outFile.ofp << "!                                                       |" << endl;
		outFile.ofp << "! ASCII Format                                          |" << endl;
		outFile.ofp << "! Native C ordering, last index varies fastest          |" << endl;
		outFile.ofp << "! Blank lines are ignored                               |" << endl;
		outFile.ofp << "! \"!\" escapes rest of line as comment                   |" << endl;
		outFile.ofp << "!-------------------------------------------------------|" << endl;
		outFile.ofp << "File_type = d" << endl;
    }
	else if(outFile.extn == ".cef"){

		// magic number 
		outFile.ofp << "!@" << outFile.extn << "@" <<  DvString(QIE_VERSION_NUM) << endl;
		outFile.ofp << "!-------- Cluster Exchange Version 2 ASCII File --------|" << endl;
		outFile.ofp << "! Free width data entries separated by delimiters       |" << endl;
		outFile.ofp << "!                                                       |" << endl;
		outFile.ofp << "! ASCII Format                                          |" << endl;
		outFile.ofp << "! Native C ordering, last index varies fastest          |" << endl;
		outFile.ofp << "! Blank lines are ignored                               |" << endl;
		outFile.ofp << "! \"!\" escapes rest of line as comment                   |" << endl;
		outFile.ofp << "!-------------------------------------------------------|" << endl;
    }
	
    if(detached){
		outFile.ofp << "FILE_NAME =" << outFile.name << ".qfh" << endl;
	}
	else{
		outFile.ofp << "FILE_NAME =" << outFile.name << outFile.extn << endl;
	}
	outFile.ofp << "!" << endl;

	if(outFile.extn != ".cef") {
		// CAA CEF 2 files are always comma delimited 
		outFile.ofp << "Attribute_delimiter = " ;
		putDelim(outFile.attr_delim);
		if(outFile.extn != ".qft"){
			outFile.ofp << "Data_delimiter = " ;
			putDelim(outFile.data_delim);
		}
	}
	
	if( outFile.extn == ".cef" && (outFile.force_eol || outFile.end_of_line != '\n' ) ){
		outFile.ofp << "End_of_record_marker = " ;
		putDelim(outFile.end_of_line, true); 
	}

	if( outFile.extn == ".qfd" && outFile.end_of_line != '\n' ){
		outFile.ofp << "End_of_record_marker = " ;
		putDelim(outFile.end_of_line, true); 
	}

	outFile.ofp << "!" << endl;

	// write global attributes

	writeGlobals();
	
	// Write Variable metadata for each variable IN ORDER 
	
	int startPosn=0;


	for(size_t n=0; n < QieVars.size(); n++){
			
		// write variable identifier object 

		writeVarBlock(QieVars[n], startPosn);
	}

    outFile.ofp << "Start_data = " << nRecords << endl;

}


// -------------------------------------------

void QieData::writeVarBlock(QieVariable *V, int &startPosn)
{	
	if(V->Vobj.is_nil()) return;
	
	outFile.ofp << "START_VARIABLE = " << V->name << endl;

	int n_items = 1;
	DvString size_str = "";
	
	if(V->Vobj->arraySize() > 1 ){
		size_str = "SIZES = ";
		int M = V->Vobj->nDims();
		for (int m=0; m<M; m++){
			size_str += (int) V->Vobj->dims[m];
			if (m < M-1) {
				size_str += outFile.attr_delim;
				size_str += " ";
			}
			else size_str += '\n';
			
			n_items *= V->Vobj->dims[m];
		}
	} // end if on num dimensions
	
	if( !size_str.empty() )  outFile.ofp << size_str;

	int StrLen = V->Vobj->maxStrLen();

	if(outFile.extn == ".qft"){
	
		// find column widths and starts

		// ensure title line fits
		V->col_width =  (V->name.size() / n_items) + 1 ;
		V->col_start =  startPosn;

		if(V->Vobj->is_dbl()){
			if(V->col_width < Q_DOUBLE_WIDTH ) V->col_width = Q_DOUBLE_WIDTH;
		}
		else if(V->Vobj->is_int()){
			if(V->col_width < Q_INT_WIDTH ) V->col_width = Q_INT_WIDTH;
		}
		else if(V->Vobj->is_str()){
			if(V->col_width <= StrLen ) V->col_width = V->Vobj->maxStrLen()+1;
		}
		else if(V->Vobj->is_time()){
			if(V->col_width <= Q_TIME_WIDTH ) V->col_width = Q_TIME_WIDTH;
		}
		else if(V->Vobj->is_event()){
			if(V->col_width <= Q_EVENT_WIDTH ) V->col_width = Q_EVENT_WIDTH;
		}
		
		if( V->recVary && V->Vobj->seqSize() > 1){
			outFile.ofp << "Column_start = " <<  V->col_start << endl;
			outFile.ofp << "Column_width = " << V->col_width << endl;
			
			// increment total 

			startPosn += V->col_width * n_items; // this gets returned, so is available to next var
		}
	}
	
	DvString dType = V->Vobj->getTypeAsText();
	if(V->datatype != DATA_STR) dType = DataTypeToStr(V->datatype); // read has set this
	
	outFile.ofp << "VALUE_TYPE = " << dType << endl;

	if(outFile.extn != ".cef" && V->Vobj->is_str()){
          outFile.ofp << "MAXSTRLEN = " << StrLen << endl;
	}

	// add data for rec_vary == NOVARY  
	if( !V->recVary ){
          outFile.ofp << "DATA = ";
          writeVarData(V, 0, outFile.attr_delim); // attr_delim as written in metadata block
          outFile.ofp << "\n";
     }

	writeVattrs(V);
   
	outFile.ofp << "END_VARIABLE = " <<  V->name << endl;
	outFile.ofp << "\n";

}

// -------------------------------------------

void QieData::writeData()
{
	char delim = outFile.data_delim;
	char var_end = ' ';
	if(outFile.end_of_line != '\n') var_end = '\n';
	
	if(QieVars.size() < 1) return;
		

	for(size_t r=0; r<nRecords; r++){
		
		writeVarData(QieVars[0], r, delim);

		for(size_t n=1; n<QieVars.size(); n++) {
		
			outFile.ofp << delim << var_end;
			writeVarData(QieVars[n], r, delim);
		}
		outFile.ofp << outFile.end_of_line;
		if(outFile.end_of_line != '\n') outFile.ofp << endl;
	}
	outFile.ofp << endl; // ensure buffer is flushed
}

// -------------------------------------------

void QieData::writeDataCAA()
{	
	if(outFile.end_of_line != '$') return writeData(); // like CEF but allows other record end
	
	if(QieVars.size() < 1) return;

	// write records
	for(size_t r=0; r<nRecords; r++){
	
		// write each variable at this rec
		writeVarDataCAA(QieVars[0], r);

		for(size_t n=1; n<QieVars.size(); n++) {
			outFile.ofp << ",\n";
			writeVarDataCAA(QieVars[n], r);
		}
		outFile.ofp << outFile.end_of_line << endl;
	}
}

// -------------------------------------------

void QieData::writeDataQFT()
{	
	if(QieVars.size() < 1) return;
	
	// write title
	QieVariable* last = QieVars[QieVars.size()-1];
	size_t len = last->col_start + last->col_width*last->Vobj->arraySize();
	
	DvString title;
	title.assign(len+1, ' ');
	for(size_t n=0; n<QieVars.size(); n++){
		QieVariable* next = QieVars[n];
		title.string::replace(next->col_start+1, next->name.size(), (string) next->name); 
		// ensure names are separated: truncates long names
		title.string::replace(next->col_start + next->col_width*next->Vobj->arraySize(), 1, " "); 
	}
	title[0]='!';
	if(title.size() < 240) outFile.ofp << title << endl;
	
	
	// write records
	for(size_t r=0; r<nRecords; r++){
		
		writeVarDataQFT(QieVars[0], r);

		for(size_t n=1; n<QieVars.size(); n++) {
			writeVarDataQFT(QieVars[n], r);
		}
		outFile.ofp << endl;
	}
}


// -------------------------------------------

void QieData::writeVarData(QieVariable *V, size_t r, char delim)
{	
	size_t len = V->Vobj->arraySize();
	size_t nstart = len*r;

	if(V->Vobj->is_dbl() && !precIsSet){
	
		outFile.ofp << V->Vobj->dbl(nstart);
		
		for(size_t i=1; i<len; i++){
			outFile.ofp << delim << " ";
			outFile.ofp << V->Vobj->dbl(nstart + i); 
		}
	}
	else if(V->Vobj->is_time()) {

		outFile.ofp << V->Vobj->time(nstart).getISOstring();

		for(size_t i=1; i<len; i++){
			outFile.ofp << delim << " ";
			outFile.ofp << V->Vobj->time(nstart + i).getISOstring();
		}
	}
	else {

		outFile.ofp << V->Vobj->asStr(nstart); // first element

		for(size_t i=1; i<len; i++){
			outFile.ofp << delim << " ";
			outFile.ofp << V->Vobj->asStr(nstart + i); 
		}
	}

}	

// -------------------------------------------

void QieData::writeVarDataCAA(QieVariable *V, size_t r)
{	
	size_t len = V->Vobj->arraySize();
	size_t nstart = len*r;


	if(V->Vobj->is_dbl() && !precIsSet){
	
		outFile.ofp << V->Vobj->dbl(nstart);
		
		for(size_t i=1; i<len; i++){
			outFile.ofp  << ",";
			outFile.ofp << V->Vobj->dbl(nstart + i); 
		}
	}
	else if(V->Vobj->is_time()) {

		outFile.ofp << V->Vobj->time(nstart).getISOstring();

		for(size_t i=1; i<len; i++){
			outFile.ofp << ",";
			outFile.ofp << V->Vobj->time(nstart + i).getISOstring();
		}
	}
	else {

		outFile.ofp << V->Vobj->asText(nstart);

		for(size_t i=1; i<len; i++){
			outFile.ofp << ",";
			outFile.ofp << V->Vobj->asText(nstart + i); // asText puts quotes round str types
		}
	}
}	

// -------------------------------------------

void QieData::writeVarDataQFT(QieVariable *V, size_t r)
{	
	size_t len = V->Vobj->arraySize();
	DvString data;
	data.assign(len * V->col_width, ' ');
	
	size_t posn=0;
	for(size_t i=0; i<len; i++){
		DvString entry = V->Vobj->asStr(r*len + i);
		data.string::replace(posn, entry.size(), entry); 
		posn += V->col_width;
	}

	outFile.ofp << data;

}	

// -------------------------------------------

void QieData::writeVattrs(QieVariable *V)
{
	int lineTotal;
	bool isCEF2 = false;
	if(outFile.extn == ".CEF") isCEF2 = true;

	DvNode *xref = V->Vobj->first_xref(); 
	
	while(xref)
	{ 
		if (xref->obj().is_nil() ) {
			xref = xref->next;
			continue;
		}
				
		outFile.ofp << xref->name() << " = ";

		if( xref->obj()->is_dbl() ){
			lineTotal = 0;
			outFile.ofp << xref->obj()->dbl(0);
			
			for (size_t m=1; m<xref->obj()->totalSize(); m++){
				lineTotal++;
				outFile.ofp << outFile.attr_delim;
				if(isCEF2 && lineTotal*(Q_DOUBLE_WIDTH+1) > MAX_LINE_LEN_CEF2 ) {
					outFile.ofp << "\\ \n";
					lineTotal = 0;
				}
				outFile.ofp << xref->obj()->dbl(m);
			}
		}
		else if( xref->obj()->is_int() ){
			lineTotal = 0;
			outFile.ofp << xref->obj()->itg(0);
			for (size_t m=1; m<xref->obj()->totalSize(); m++){
				lineTotal++;
				outFile.ofp << outFile.attr_delim;
				if(isCEF2 && lineTotal*(Q_INT_WIDTH+1) > MAX_LINE_LEN_CEF2 ) {
					outFile.ofp << "\\ \n";
					lineTotal = 0;
				}
				outFile.ofp << xref->obj()->itg(m);
			}
		}
		else if( xref->obj()->is_str() ){
			if(isCEF2) outFile.ofp << "\"" << xref->obj()->str(0) << "\"";
			else outFile.ofp << xref->obj()->str(0);
			if(isCEF2) lineTotal = xref->obj()->str(0).size()+1;
			else lineTotal = xref->obj()->str(0).size()+3;
			
			for (size_t m=1; m<xref->obj()->totalSize(); m++){
			
				if(isCEF2) lineTotal += xref->obj()->str(m).size()+1;
				else lineTotal += xref->obj()->str(m).size()+3;

				outFile.ofp << outFile.attr_delim;
				outFile.ofp << " ";
//				if(isCEF2 && lineTotal > MAX_LINE_LEN_CEF2 ) {
//					outFile.ofp << "\\ \n";
//					lineTotal = 0;
//				}
				if(isCEF2) outFile.ofp << "\"" << xref->obj()->str(m) << "\"";
				else outFile.ofp << xref->obj()->str(m);
			}
		}
		else if( xref->obj()->is_time() ){
			// set resolution in DvTime library using any DvTime object
			int oldSecRes = xref->obj()->time(0).setSecResolution(xref->obj()->secRes);
			
			DvString Tstr = xref->obj()->time(0).getISOstring();
			outFile.ofp << Tstr;
			lineTotal = Tstr.size();
			
			for (size_t m=1; m<xref->obj()->totalSize(); m++){
				Tstr = xref->obj()->time(m).getISOstring();
				
				lineTotal += Tstr.size()+1;
				outFile.ofp << outFile.attr_delim;
				if(isCEF2 && lineTotal > MAX_LINE_LEN_CEF2 ) {
					outFile.ofp << "\\ \n";
					lineTotal = 0;
				}
				outFile.ofp << Tstr;
			}
			// reset time resolution
			xref->obj()->time(0).setSecResolution(oldSecRes);
		}
		else if( xref->obj()->is_event() ){
			// set resolution in DvTime library using any DvTime object
			int oldSecRes = xref->obj()->event(0).start().setSecResolution(xref->obj()->secRes);
			
			DvString Estr = xref->obj()->event(0).getISOstring();
			outFile.ofp << Estr;
			lineTotal = Estr.size();
			
			for (size_t m=1; m<xref->obj()->totalSize(); m++){
				Estr = xref->obj()->event(m).getISOstring();
				
				lineTotal += Estr.size()+1;
				outFile.ofp << outFile.attr_delim;
				if(isCEF2 && lineTotal > MAX_LINE_LEN_CEF2 ) {
					outFile.ofp << "\\ \n";
					lineTotal = 0;
				}
				outFile.ofp << Estr;
			}
			// reset time resolution
			xref->obj()->event(0).start().setSecResolution(oldSecRes);
		}
       
		outFile.ofp << endl;
		xref = xref->next;
		
	} // end while 

}	

// -------------------------------------------

void QieData::setSecDecimals(int value){

	dataSecRes = value;
   
}

// -------------------------------------------

void QieData::writeGlobals()
{		
	DvNode *Gnode = GlobalList.first();
	while (Gnode){
		int oldRes=3;
		int n_entries = Gnode->obj()->seqSize();
		
		outFile.ofp << "Start_meta = " << Gnode->name() << endl;
		
		if(outFile.extn != ".cef") {
			if(Gnode->name() == "Software_version") outFile.ofp << "Number_of_entries = " << n_entries+1 << endl;
			else  outFile.ofp << "Number_of_entries = " << n_entries << endl;
		}

		DvString dType = Gnode->obj()->getTypeAsText();
		
	    outFile.ofp << "VALUE_TYPE = " << dType << endl;
		
		if(Gnode->obj()->is_time()) {
			oldRes = Gnode->obj()->time(0).setSecResolution(Gnode->obj()->secRes);
		}
		
		for (int r=0; r < n_entries; r++){

			// write entry
			writeGEntry( Gnode->obj(), r);

		}

		if(Gnode->obj()->is_time()) {
			Gnode->obj()->time(0).setSecResolution(oldRes);
		}
		
		if(Gnode->name() == "Software_version"){
			if(outFile.extn == ".cef")	outFile.ofp << "Entry = \"" <<  DvString(QIE_VERSION_NUM) << "\"" << endl;
			else				outFile.ofp << "Entry = " << DvString(QIE_VERSION_NUM) << endl;
		}
		
		outFile.ofp << "End_meta = " << Gnode->name() << endl;
		outFile.ofp << "!" << endl;
		
		Gnode = Gnode->next;
	}
}

// -------------------------------------------

void QieData::writeGEntry(DvObject_var &dobj, int r)
{
	outFile.ofp << "Entry = ";
	size_t len = dobj->arraySize();
	if(outFile.extn == ".CEF"){
		for(size_t i=0; i<len; i++){
			outFile.ofp << dobj->asText(r*len + i); // quoted strings
			if(i < len-1 ) outFile.ofp << ", ";
			else outFile.ofp << endl;
		}
	}
	else{
		for(size_t i=0; i<len; i++){
			outFile.ofp << dobj->asStr(r*len + i);
			if(i < len-1 ) outFile.ofp << ", ";
			else outFile.ofp << endl;
		}
	}
}

// -------------------------------------------

DvString QieData::addObj(DvString name, DvObject_var obj, int accuracy){ 
	// used by QSAS export
	
	int Vnum = findVarNumByName(name.c_str());
	if(Vnum != -1) {
		if( !this->QieVars[Vnum]->Vobj->same(*obj) ){
			// same name but new object, find new name
			int i=1;
			DvString stem = name;
			stem += "_";
			name = stem+i;
			while(findVarNumByName(name.c_str()) != -1) {
				i++;
				name = stem+i;
			}
		}
		else return name; // already on list
	}
	
	QieVariable *V = new QieVariable(name.c_str()); 
	this->QieVars.push_back(V);
	V->Vobj = new DvObject(obj); // copy to avoid modifying WL object
	V->nRecs = V->Vobj->seqSize();
	if(V->nRecs > 1 || nRecords == 1) V->recVary = VARY;
	else V->recVary = NOVARY;
		
	// Find data type

	if( V->Vobj->is_dbl() ) V->datatype = DATA_DBL; 
	else if( V->Vobj->is_str() ) V->datatype = DATA_STR; 
	else if( V->Vobj->is_time() ) V->datatype = DATA_TIME; 
	else if( V->Vobj->is_event() ) V->datatype = DATA_EVENT;
	else if( V->Vobj->is_int() ) V->datatype = DATA_INT;

	V->Vobj->secRes = accuracy; // in case another var is time or event

	// convert ptr types to name
	DvNode *xref = V->Vobj->first_xref();
	while(xref){
        if(xref->obj()->is_time() && d0_var_num == -1) d0_var_num = QieVars.size();

        if(xref->obj()->xref_exists(PTR_VAR_NAME)){
			DvObject_var ptrXref = xref->obj()->get_xref(PTR_VAR_NAME);

			addObj(ptrXref->asStr(0), xref->obj(), accuracy);
			xref->setObj(ptrXref);
		}
        else if(xref->name() == DEPEND_0){
            
            // processing or join may have modified Dep_0 and lost original name
            DvString depName;
            if(xref->obj()->is_time()) depName = addObj("Timetags", xref->obj(), accuracy);
            else depName = addObj("Dependency", xref->obj(), accuracy);
            
            DvObject_var ptrXref = new DvObject(depName);
            V->Vobj->change_xref(DEPEND_0, ptrXref);
            
        }
		else if(xref->obj()->seqSize() > 1){
			// must be written as record varying
			DvString namePtr = xref->name()+"_variable";
			DvObject_var ptrXref = new DvObject(namePtr);

			addObj(namePtr, xref->obj(), accuracy);
			xref->setObj(ptrXref);
		}
		else if(xref->obj()->xref_exists(SI_CONVERSION) || xref->obj()->xref_exists(DELTA_PLUS )){
            // object with xrefs also needs to be added as object
            DvString namePtr = xref->name()+"_ptr"; 
			DvObject_var ptrXref = new DvObject(namePtr);;
            
            addObj(namePtr, xref->obj(), accuracy);
			xref->setObj(ptrXref);
		}
		xref = xref->next;
	}
    V->Vobj->delete_xref(PTR_VAR_NAME);
    
	return name;
}

// -------------------------------------------

bool  QieData::ensureFileName(bool replace, bool detached) {

	DvString extn = outFile.extn;
	DvString name = outFile.name;
	if(detached) {
		extn = ".qfh";
		if( !outFile.header.empty() ) name = outFile.header;
	}
	// ensure name exists
	if( name.empty() )  {
		name = "./qsas_data";
	}
	
	// Ensure this isn't writing to root
	outFile.name = outFile.safePath(outFile.name.c_str());
	
	DvString fullName = outFile.name + extn;
	
	// first check if file exists
	if(!replace){
		FILE *fpt=fopen(fullName.c_str(), "r");
		if( fpt != NULL ) {
			/* it exists, overwrite not allowed, bale out */
			fclose(fpt);
			return false;
		}
	}
	
	if(extn == ".cdf") return true;
	
	// open file for write 

	outFile.ofp.open(fullName.c_str(), std::fstream::out );

	if(outFile.ofp.is_open()) return true;
		
	return false;

} 
// ------------------------------------------

void QieData::putDelim (char delim, bool quoted ){
	if(quoted) outFile.ofp << "\"" ;
	if(delim == '\n') outFile.ofp << "\\n" << endl;
	else if(delim == '\t') outFile.ofp << "\\t" << endl; 
	else if(delim == '\'') outFile.ofp << "\\\'" << endl;
	else if(delim == '\"') outFile.ofp << "\\\"" << endl;
	else outFile.ofp << delim;
	
	if(quoted) outFile.ofp << "\"" ;
	outFile.ofp  << endl;

} 


// ------------------Utilities----------------
// -------------------------------------------

int getSecResolution(string &value){

	size_t posn = value.rfind('.');
	if(posn == string::npos) return 0;
	posn++;
	int len = 0;
	while( isdigit(value[posn+len]) ) len++;

	return len;
}


// -------------------------------------------



